import pytest

import prompty
from prompty.azure.processor import AzureOpenAIProcessor
from prompty.invoker import InvokerFactory
from prompty.serverless.processor import ServerlessProcessor
from tests.fake_azure_executor import FakeAzureExecutor
from tests.fake_serverless_executor import FakeServerlessExecutor


@pytest.fixture(scope="module", autouse=True)
def fake_azure_executor():
    InvokerFactory.add_executor("azure", FakeAzureExecutor)
    InvokerFactory.add_executor("azure_openai", FakeAzureExecutor)
    InvokerFactory.add_executor("azure_beta", FakeAzureExecutor)
    InvokerFactory.add_executor("azure_openai_beta", FakeAzureExecutor)
    InvokerFactory.add_processor("azure", AzureOpenAIProcessor)
    InvokerFactory.add_processor("azure_openai", AzureOpenAIProcessor)
    InvokerFactory.add_executor("azure_beta", AzureOpenAIProcessor)
    InvokerFactory.add_executor("azure_openai_beta", AzureOpenAIProcessor)
    InvokerFactory.add_executor("serverless", FakeServerlessExecutor)
    InvokerFactory.add_processor("serverless", ServerlessProcessor)


def test_basic_load():
    prompt = "tools/basic.prompty"
    p = prompty.load(prompt)
    assert len(p.tools) == 1
    assert p.tools[0].id == "bing"
    assert p.tools[0].type == "web_search"
    assert p.tools[0].description == "A tool that can search the web for information."
    assert "url" in p.tools[0].options and p.tools[0].options["url"] == "${env:BING_URL}"
    assert len(p.tools[0].parameters) == 2
    print(p)


@pytest.mark.asyncio
async def test_basic_async_load():
    prompt = "tools/basic.prompty"
    p = await prompty.load_async(prompt)
    assert len(p.tools) == 1
    assert p.tools[0].id == "bing"
    assert p.tools[0].type == "web_search"
    assert p.tools[0].description == "A tool that can search the web for information."
    assert "url" in p.tools[0].options and p.tools[0].options["url"] == "${env:BING_URL}"
    assert len(p.tools[0].parameters) == 2
    print(p)


def test_dynamic_load():
    prompt = "tools/dynamic.prompty"
    p = prompty.load(prompt)
    prompty.prepare(p, merge_sample=True)
    assert len(p.tools) == 2
    assert p.tools[0].id == "bing"
    assert p.tools[0].type == "web_search"
    assert p.tools[0].description == "A tool that can search the web for information."
    assert "url" in p.tools[0].options and p.tools[0].options["url"] == "https://api.bing.microsoft.com/v7.0/search"
    assert len(p.tools[0].parameters) == 2

    assert p.tools[1].id == "callable"
    assert p.tools[1].type == "function"
    print(p)


@pytest.mark.asyncio
async def test_dynamic_async_load():
    prompt = "tools/dynamic.prompty"
    p = await prompty.load_async(prompt)
    await prompty.prepare_async(p, merge_sample=True)
    assert len(p.tools) == 2
    assert p.tools[0].id == "bing"
    assert p.tools[0].type == "web_search"
    assert p.tools[0].description == "A tool that can search the web for information."
    assert "url" in p.tools[0].options and p.tools[0].options["url"] == "https://api.bing.microsoft.com/v7.0/search"
    assert len(p.tools[0].parameters) == 2

    assert p.tools[1].id == "callable"
    assert p.tools[1].type == "function"
    print(p)



import importlib
import json
import os
from dataclasses import asdict, is_dataclass
from pathlib import Path
from typing import Any, Optional

import click
from dotenv import load_dotenv

import prompty
from prompty.tracer import PromptyTracer, Tracer, console_tracer, trace


def normalize_path(p, create_dir=False) -> Path:
    path = Path(p)
    if not path.is_absolute():
        path = Path(os.getcwd()).joinpath(path).absolute().resolve()
    else:
        path = path.absolute().resolve()

    if create_dir:
        if not path.exists():
            print(f"Creating directory {str(path)}")
            os.makedirs(str(path))

    return path


def dynamic_import(module: str):
    # built in modules
    if module == "azure" or module == "azure_openai":
        t = "prompty.azure"
    elif module == "serverless":
        t = "prompty.serverless"
    elif module == "openai":
        t = "prompty.openai"
    else:
        t = module

    print(f"Loading invokers from {t}")
    importlib.import_module(t)


@trace
def chat_mode(prompt_path: str):
    W = "\033[0m"  # white (normal)
    R = "\033[31m"  # red
    G = "\033[32m"  # green
    # O = "\033[33m"  # orange
    B = "\033[34m"  # blue
    # P = "\033[35m"  # purple
    print(f"Executing {str(prompt_path)} in chat mode...")
    p = prompty.load(str(prompt_path))
    if p.get_input("thread") is None:
        print(f"{R}{str(prompt_path)} needs to have a thread input to work in chat mode{W}")
        return

    if p.get_input("query") is None:
        print(f"{R}{str(prompt_path)} needs to have a query input to work in chat mode{W}")
        return

    else:

        try:
            # load executor / processor types
            dynamic_import(p.model.connection["type"])
            while True:
                user_input = input(f"\n{B}User:{W} ")
                if user_input == "exit":
                    break
                # reloadable prompty file
                result = prompty.execute(p, inputs={"query": user_input}, merge_sample=True)
                print(f"\n{G}Assistant:{W} {result}")
        except Exception as e:
            print(f"{type(e).__qualname__}: {e}")

    print(f"\n{R}Goodbye!{W}\n")


def execute(prompt_path: str, inputs: Optional[dict[str, Any]] = None, raw=False):
    name = Path(prompt_path).name.lower().replace(" ", "-").replace(".prompty", "")
    inputs = inputs or {}
    with Tracer.start(name) as trace:
        trace("type", "cli")
        trace("signature", "prompty.cli.execute")
        trace("description", "Prompty CLI Execution")
        trace("inputs", {"prompt_path": prompt_path, "inputs": inputs})

        p = prompty.load(prompt_path)

        try:
            # load executor / processor types
            dynamic_import(p.model.connection["type"])

            result = prompty.execute(p, inputs=inputs, raw=raw, merge_sample=True)
            trace("result", result)
            if is_dataclass(result) and not isinstance(result, type):
                print("\n", json.dumps(asdict(result), indent=4), "\n")
            elif isinstance(result, list):
                print("\n", json.dumps([asdict(item) for item in result], indent=4), "\n")
            else:
                print("\n", result, "\n")
        except Exception as e:
            print(f"{type(e).__qualname__}: {e}", "\n")


def _attributes_to_dict(ctx: click.Context, attribute: click.Option, attributes: tuple[str, ...]) -> dict[str, str]:
    """Click callback that converts attributes specified in the form `key=value` to a
    dictionary"""
    result = {}
    for arg in attributes:
        k, v = arg.split("=")
        if k in result:
            raise click.BadParameter(f"Attribute {k!r} is specified twice")
        if v == "@-":
            v = click.get_text_stream("stdin").read()
        if v.startswith("@"):
            v = Path(v[1:]).read_text()
        result[k] = v

    return result


@click.command(
    epilog="""
\b
INPUTS: key=value pairs
    The values can come from:
    - plain strings - e.g.: question="Does it have windows?"
    - files - e.g.: question=@question.txt
    - stdin - e.g.: question=@-

For more information, visit https://prompty.ai/
"""
)
@click.option("--source", "-s", required=True)
@click.option("--env", "-e", required=False)
@click.option("--verbose", "-v", is_flag=True)
@click.option("--chat", "-c", is_flag=True)
@click.argument("inputs", nargs=-1, callback=_attributes_to_dict)
@click.version_option()
def run(source, env, verbose, chat, inputs):
    # load external env file
    if env:
        print(f"Loading environment variables from {env}")
        load_dotenv(env)

    prompt_path = normalize_path(source)
    if not prompt_path.exists():
        print(f"{str(prompt_path)} does not exist")
        return

    if verbose:
        Tracer.add("console", console_tracer)

    ptrace = PromptyTracer()
    Tracer.add("prompty", ptrace.tracer)

    if chat:
        chat_mode(str(prompt_path))
    else:
        execute(str(prompt_path), inputs=inputs)


if __name__ == "__main__":
    chat_mode(source="./tests/prompts/basic.prompt")



import json
import typing
from pathlib import Path

from openai.types.chat import ChatCompletionChunk
from openai.types.chat.chat_completion import ChatCompletion
from openai.types.create_embedding_response import CreateEmbeddingResponse

from prompty import Prompty
from prompty.core import AsyncPromptyStream, PromptyStream
from prompty.invoker import Invoker


## Azure Fake Executor
## To save on OpenAI Calls, will used known
## cached responses using invoker pattern
class FakeAzureExecutor(Invoker):
    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)
        self.resolve_model()
        self.api = self.prompty.model.api
        self.deployment = self.prompty.model.connection["azure_deployment"]
        self.options = self.prompty.model.options

    def invoke(self, data: typing.Any) -> typing.Any:
        if self.prompty.file:
            if isinstance(self.prompty.file, str):
                self.prompty.file = Path(self.prompty.file).resolve().absolute()

            p = Path(self.prompty.file.parent) / f"{self.prompty.file.name}.execution.json"
            with open(p, encoding="utf-8") as f:
                j = f.read()

            if self.options.get("stream", False):
                items = json.loads(j)

                def generator():
                    for i in range(1, len(items)):
                        yield ChatCompletionChunk.model_validate(items[i])

                return PromptyStream("FakeAzureExecutor", generator())

            elif self.api == "chat":
                return ChatCompletion.model_validate_json(j)
            elif self.api == "embedding":
                return CreateEmbeddingResponse.model_validate_json(j)

        elif self.api == "embedding":
            if not isinstance(data, list):
                d = [data]
            else:
                d = data

            n = "-".join([s.replace(" ", "_") for s in d if isinstance(s, str)])
            p = Path(__file__).parent / f"{n}.embedding.json"
            with open(p, encoding="utf-8") as f:
                response = CreateEmbeddingResponse.model_validate_json(f.read())
                return response

        return data

    async def invoke_async(self, data: str) -> typing.Any:
        """Invoke the Prompty Chat Parser (Async)

        Parameters
        ----------
        data : str
            The data to parse

        Returns
        -------
        str
            The parsed data
        """
        if self.prompty.file:
            if isinstance(self.prompty.file, str):
                self.prompty.file = Path(self.prompty.file).resolve().absolute()
            p = Path(self.prompty.file.parent) / f"{self.prompty.file.name}.execution.json"
            with open(p, encoding="utf-8") as f:
                j = f.read()

            if self.options.get("stream", False):
                items = json.loads(j)

                async def generator():
                    for i in range(1, len(items)):
                        yield ChatCompletionChunk.model_validate(items[i])

                return AsyncPromptyStream("FakeAzureExecutor", generator())

            elif self.api == "chat":
                return ChatCompletion.model_validate_json(j)
            elif self.api == "embedding":
                return CreateEmbeddingResponse.model_validate_json(j)

        elif self.api == "embedding":
            if not isinstance(data, list):
                d = [data]
            else:
                d = data

            n = "-".join([s.replace(" ", "_") for s in d if isinstance(s, str)])
            p = Path(__file__).parent / f"{n}.embedding.json"
            with open(p, encoding="utf-8") as f:
                response = CreateEmbeddingResponse.model_validate_json(f.read())
                return response

        return data



# __init__.py
from prompty.invoker import InvokerException

try:
    # Reuse the common Azure OpenAI Processor
    from ..azure.processor import AzureOpenAIProcessor  # noqa
    from .executor import AzureOpenAIBetaExecutor  # noqa
except ImportError:
    raise InvokerException("Error registering AzureOpenAIBetaExecutor and AzureOpenAIProcessor", "azure_beta")



import copy
import typing
import uuid
import warnings
from collections.abc import AsyncIterator, Iterator
from dataclasses import asdict, dataclass, field, fields
from pathlib import Path
from typing import Callable, Literal, Optional, Union

from .tracer import Tracer, to_dict
from .utils import get_json_type, load_json, load_json_async


@dataclass
class ToolCall:
    id: str
    name: str
    arguments: str


@dataclass
class InputProperty:
    """InputProperty class to define the input properties of the model

    Attributes
    ----------
    type : str
        The type of the property
    description : str
        The description of the property
    required : bool
        Whether the property is required or not
    default : any
        The default value of the property
    sample : any
        The sample value of the property
    strict : bool
        Whether the property is strict or not
    json_schema : dict
        The json schema of the property (optional)
    """

    type: Literal["string", "number", "array", "object", "boolean"]
    name: str = field(default="")
    description: str = field(default="")
    required: bool = field(default=False)
    default: typing.Any = field(default=None)
    sample: typing.Any = field(default=None)
    strict: bool = field(default=True)
    # used internally
    value: typing.Any = field(default=None)
    json_schema: Optional[dict] = field(default_factory=dict)


@dataclass
class OutputProperty:
    """OutputProperty class to define the output properties of the model

    Attributes
    ----------
    type : str
        The type of the property
    description : str
        The description of the property
    default : any
        The default value of the property
    json_schema : dict
        The json schema of the property (optional)
    """

    type: Literal["string", "number", "array", "object", "boolean"]
    name: str = field(default="")
    description: str = field(default="")
    # only for array type - need to update schema
    items: list["OutputProperty"] = field(default_factory=list)

    json_schema: Optional[dict] = field(default_factory=dict)


@dataclass
class ModelProperty:
    """ModelSettings class to define the model of the prompty

    Attributes
    ----------
    api : str
        The api of the model
    connection : dict
        The connection of the model
    parameters : dict
        The parameters of the model
    response : dict
        The response of the model
    """

    id: str = field(default="")
    api: str = field(default="")
    connection: dict = field(default_factory=dict)
    options: dict = field(default_factory=dict)


@dataclass
class TemplateProperty:
    """TemplateSettings class to define the template of the prompty

    Attributes
    ----------
    type : str
        The type of the template
    parser : str
        The parser of the template
    nonce : str
        Nonce is automatically genereted for each run
    content : str
        Template content used for rendering
    """

    format: str = field(default="mustache")
    parser: str = field(default="")
    nonce: str = field(default="")
    content: Union[str, list[str], dict] = field(default="")
    strict: bool = field(default=False)
    options: dict = field(default_factory=dict)


@dataclass
class ToolParameter:
    name: str
    type: Literal["string", "number", "array", "object", "boolean"]
    description: str = field(default="")
    required: bool = field(default=False)
    enum: list[typing.Any] = field(default_factory=list)


@dataclass
class ToolProperty:
    id: str
    type: str
    description: Optional[str] = field(default="")
    options: dict[str, typing.Any] = field(default_factory=dict)
    # used internally for assigning funcs
    value: typing.Any = field(default=None)
    parameters: list[ToolParameter] = field(default_factory=list)


@dataclass
class Prompty:
    """Prompty class to define the prompty

    Attributes
    ----------
    id : str
        The id of the prompty
    version : str
        The version of the prompty
    name : str
        The name of the prompty
    description : str
        The description of the prompty
    metadata : dict
        The metadata of the prompty
    base : str
        The base of the prompty
    basePrompty : Prompty
        The base prompty
    model : ModelSettings
        The model of the prompty
    inputs : dict[str, PropertySettings]
        The inputs of the prompty
    outputs : dict[str, PropertySettings]
        The outputs of the prompty
    template : TemplateSettings
        The template of the prompty
    tools:
        The tools of the prompty
    file : FilePath
        The file of the prompty
    content : Union[str, list[str], dict]
        The content of the prompty
    """

    # metadata
    id: str = field(default=uuid.uuid4().hex)
    version: str = field(default="")
    name: str = field(default="")
    description: str = field(default="")
    metadata: dict[str, typing.Any] = field(default_factory=dict)

    # future: undocumented template inheritance
    base: str = field(default="")
    basePrompty: Union["Prompty", None] = field(default=None)

    # model execution
    model: ModelProperty = field(default_factory=ModelProperty)

    # input / output
    inputs: list[InputProperty] = field(default_factory=list)
    outputs: list[OutputProperty] = field(default_factory=list)

    # tools
    tools: list[ToolProperty] = field(default_factory=list)

    # template
    template: TemplateProperty = field(default_factory=TemplateProperty)

    file: Union[str, Path] = field(default="")
    instructions: str = field(default="")
    additional_instructions: str = field(default="")
    content: Union[str, list[str], dict] = field(default="")
    instructions = field(default="")
    additional_instructions = field(default="")

    slots: list[dict[str, str]] = field(default_factory=list)

    def get_input(self, name: str) -> Union[InputProperty, None]:
        """Get the property of the prompty

        Parameters
        ----------
        name : str
            The name of the property

        Returns
        -------
        InputProperty
            The property of the prompty
        """
        for i in self.inputs:
            if i.name == name:
                return i
        return None

    def get_tool(self, name: str) -> Union[ToolProperty, None]:
        """Get the property of the prompty

        Parameters
        ----------
        name : str
            The name of the property

        Returns
        -------
        ToolProperty
            The property of the prompty
        """
        for i in self.tools:
            if i.id == name:
                return i
        return None

    def set_input_value(self, name: str, value: typing.Any) -> None:
        """Set the value of the input property"""
        for i in self.inputs:
            if i.name == name:
                i.value = value
                return

        raise ValueError(f"Input {name} not found")

    def set_tool_value(self, name: str, value: typing.Any) -> None:
        """Set the value of the input property"""
        for i in self.tools:
            if i.id == name:
                i.value = value
                return

        raise ValueError(f"Tool {name} not found")

    def get_output(self, name: str) -> OutputProperty:
        """Get the output property of the prompty

        Parameters
        ----------
        name : str
            The name of the property

        Returns
        -------
        OutputProperty
            The property of the prompty
        """

        for i in self.outputs:
            if i.name == name:
                return i
        raise ValueError(f"Property {name} not found")

    def to_safe_dict(self) -> dict[str, typing.Any]:
        d: dict[str, typing.Any] = {}
        for items in fields(self):
            k = items.name
            v = getattr(self, k)
            if v != "" and v != {} and v != [] and v is not None:
                if k == "model":
                    d[k] = asdict(self.model)
                elif k == "template":
                    d[k] = asdict(self.template)
                elif k == "inputs" or k == "outputs":
                    d[k] = copy.deepcopy(v)
                elif k == "file":
                    d[k] = str(self.file.as_posix()) if isinstance(self.file, Path) else self.file
                elif k == "tools":
                    d[k] = [asdict(t) for t in v]
                elif k == "basePrompty":
                    # no need to serialize basePrompty
                    continue

                else:
                    d[k] = v
        return d

    def get_sample(self) -> dict[str, typing.Any]:
        sample = {}
        for input in self.inputs:
            if input.sample:
                sample[input.name] = input.sample
            elif input.default:
                sample[input.name] = input.default
        return sample

    def merge_tools(self, tools: list[ToolProperty]) -> None:
        self.tools = [*self.tools, *tools]

    @staticmethod
    def load_tools(tools: list[dict]) -> list[ToolProperty]:
        loaded_tools = []

        for t in tools:
            options = t.pop("options") if "options" in t else {}
            params = t.pop("parameters") if "parameters" in t else []
            parameters: dict[str, ToolParameter] = {}
            if isinstance(params, dict):
                # if parameters is a dict, convert to list of ToolParameter
                parameters = {k: ToolParameter(name=k, **v) for k, v in params.items()}
            elif isinstance(params, list):
                # if parameters is a list, convert to list of ToolParameter
                parameters = {p["name"]: ToolParameter(**p) for p in params}
            elif params:
                raise ValueError("Parameters must be a list or dict")

            # hoist params from config if they exist
            if "parameters" in options:
                params = options.pop("parameters")
                if isinstance(params, dict):
                    for k, v in params.items():
                        if k not in parameters:
                            parameters[k] = ToolParameter(name=k, **v)
                        else:
                            raise ValueError(f"Duplicate parameter {k} in options and parameters")
                elif isinstance(params, list):
                    for p in params:
                        if p["name"] not in parameters:
                            parameters[p["name"]] = ToolParameter(**p)
                        else:
                            raise ValueError(f"Duplicate parameter {p['name']} in configuration and parameters")
                else:
                    raise ValueError("Parameters must be a list or dict")

            if t["type"] == "function" and parameters == {}:
                # if function, need to have parameters
                raise ValueError("Function tools must have parameters")

            loaded_tools.append(ToolProperty(**t, parameters=[*parameters.values()], options=options))

        return loaded_tools

    @staticmethod
    def load_property(
        value: typing.Any,
        cls: type[typing.Any],
        default: Union[Callable[[], dict[str, typing.Any]], None] = None,
    ) -> dict[str, typing.Any]:

        if isinstance(value, dict) and any([f.name in value for f in fields(cls)]):
            return {**value}
        else:
            if default is not None:
                return {**default()}
            else:
                raise ValueError(f"{cls.__name__} parameters mismatch")

    @staticmethod
    def load_collection_property(
        values: Union[dict, list],
        materialize: Callable[[str, typing.Any], dict[str, typing.Any]],
        error_message: str = "",
    ) -> list[dict[str, typing.Any]]:
        if isinstance(values, list):
            return [materialize("", v) for v in values]
        elif isinstance(values, dict):
            return [materialize(k, v) for k, v in values.items()]
        else:
            if error_message == "":
                error_message = f"Collection must be a list or dict, got {type(values)}"
            else:
                error_message = f"{error_message}, got {type(values)}"
            raise ValueError(error_message)

    @staticmethod
    def load_input_property(name: str, value: typing.Any) -> dict[str, typing.Any]:

        def to_dict() -> dict[str, typing.Any]:
            return {
                "type": get_json_type(type(value)),
                "sample": value,
                "name": name,
            }

        prop = Prompty.load_property(value, InputProperty, to_dict)
        if "name" not in prop or prop["name"] == "":
            prop["name"] = name

        return prop

    @staticmethod
    def load_output_property(name: str, value: typing.Any) -> dict[str, typing.Any]:

        def to_dict() -> dict[str, typing.Any]:
            return {
                "type": get_json_type(type(value)),
                "name": name,
            }

        prop = Prompty.load_property(value, OutputProperty, to_dict)
        if "name" not in prop or prop["name"] == "":
            prop["name"] = name

        return prop

    @staticmethod
    def load_tool_param(name: str, value: typing.Any) -> dict[str, typing.Any]:

        prop = Prompty.load_property(value, ToolParameter)
        if "name" not in prop or prop["name"] == "":
            prop["name"] = name

        return prop

    @staticmethod
    def load_tool(id: str, tool: dict[str, typing.Any]) -> dict[str, typing.Any]:
        options = tool.pop("options") if "options" in tool else {}
        parameters = Prompty.load_collection_property(
            tool.pop("parameters") if "parameters" in tool else [],
            Prompty.load_tool_param,
            "Tool parameters must be a list or dict",
        )

        if "parameters" in options:
            params = Prompty.load_collection_property(
                options.pop("parameters"),
                Prompty.load_tool_param,
                "Tool parameters must be a list or dict",
            )
            parameters = [*parameters, *params]

        t = {
            **tool,
            "options": options,
            "parameters": parameters,
        }

        if "id" not in t:
            t["id"] = id

        return t

    @staticmethod
    def load_manifest(attributes: dict, content: str, global_config: dict) -> dict[str, typing.Any]:

        prompty: dict[str, typing.Any] = {}

        if "base" in attributes:
            attributes.pop("base")
            warnings.warn("base prompty is currently not supported", DeprecationWarning)

        if "model" not in attributes:
            prompty["model"] = {}

        model_props = attributes.pop("model")
        if "configuration" in model_props:
            warnings.warn(
                "Model configuration is deprecated, use connection instead",
                DeprecationWarning,
            )
            model_props["connection"] = model_props.pop("configuration")

        if "parameters" in model_props:
            warnings.warn(
                "Model parameters is deprecated, use options instead",
                DeprecationWarning,
            )
            model_props["options"] = model_props.pop("parameters")

        # load connection settings
        if "connection" not in model_props:
            model_props["connection"] = global_config
        else:
            model_props["connection"] = param_hoisting(
                model_props["connection"],
                global_config,
            )

        prompty["model"] = model_props

        # pull template settings

        if "template" in attributes:
            t = attributes.pop("template")
            if "type" in t:
                warnings.warn(
                    "Template type is deprecated, use format instead",
                    DeprecationWarning,
                )
                t["format"] = t.pop("type")

            if isinstance(t, dict):
                prompty["template"] = t
            # has to be a string denoting the type
            else:
                prompty["template"] = {"format": t, "parser": "prompty"}
        else:
            prompty["template"] = {"format": "jinja2", "parser": "prompty"}

        # formalize inputs and outputs
        inputs: list[dict[str, typing.Any]] = []
        if "inputs" in attributes:
            raw_inputs = attributes.pop("inputs")
            inputs = Prompty.load_collection_property(
                raw_inputs, Prompty.load_input_property, "Inputs must be a list or dict"
            )

        prompty["inputs"] = inputs

        outputs: list[dict[str, typing.Any]] = []
        if "outputs" in attributes:
            raw_outputs = attributes.pop("outputs")
            outputs = Prompty.load_collection_property(
                raw_outputs,
                Prompty.load_output_property,
                "Outputs must be a list or dict",
            )

        prompty["outputs"] = outputs

        tools = []
        if "tools" in attributes:
            tools_attribute = attributes.pop("tools")
            tools = Prompty.load_collection_property(tools_attribute, Prompty.load_tool, "Tools must be a list or dict")

        prompty["tools"] = tools

        if "sample" in attributes:
            raise ValueError("Sample is deprecated, use inputs instead")

        metadata = attributes.pop("metadata") if "metadata" in attributes else {}
        # DEPRECATED: authors and tags now in metadata
        if "authors" in attributes:
            warnings.warn(
                "Authors is deprecated, add authors to metadata instead",
                DeprecationWarning,
            )
            authors = attributes.pop("authors")
            if isinstance(authors, list):
                metadata["authors"] = authors
            else:
                raise ValueError("Authors must be a list")
        if "tags" in attributes:
            warnings.warn("Tags is deprecated, add tags to metadata instead", DeprecationWarning)
            tags = attributes.pop("tags")
            if isinstance(tags, list):
                metadata["tags"] = tags
            else:
                raise ValueError("Tags must be a list")

        prompty["metadata"] = metadata

        if isinstance(content, str):
            if "\n![thread]" in str(content):
                # if the content contains a thread, split it
                # into instructions and additional instructions
                instructions = content.split("\n![thread]")
                attributes["instructions"] = instructions[0]
                attributes["additional_instructions"] = instructions[1]
                # add thread input if it does not exist
                prompty["inputs"].append(
                    {"type": "array", "name": "thread", "description": "Agent Thread", "default": [], "strict": False}
                )
            else:
                attributes["instructions"] = content
                attributes["additional_instructions"] = ""

        return {**attributes, **prompty, "content": content}

    @staticmethod
    def load_raw(attributes: dict, file: Path) -> "Prompty":

        prompty = Prompty(
            model=ModelProperty(**attributes.pop("model")),
            inputs=[InputProperty(**i) for i in attributes.pop("inputs", [])],
            outputs=[OutputProperty(**i) for i in attributes.pop("outputs", [])],
            tools=Prompty.load_tools(attributes.pop("tools", [])),
            template=TemplateProperty(**attributes.pop("template")),
            file=file,
            **attributes,
        )

        # setting template scratch pad
        prompty.template.content = prompty.content
        return prompty

    @staticmethod
    def hoist_base_prompty(top: "Prompty", base: "Prompty") -> "Prompty":
        top.name = base.name if top.name == "" else top.name
        top.description = base.description if top.description == "" else top.description
        top.metadata = param_hoisting(top.metadata, base.metadata)
        top.version = base.version if top.version == "" else top.version

        top.model.api = base.model.api if top.model.api == "" else top.model.api
        top.model.connection = param_hoisting(top.model.connection, base.model.connection)
        top.model.options = param_hoisting(top.model.options, base.model.options)

        top.basePrompty = base

        # TODO: Hoist tools

        return top

    @staticmethod
    def _process_file(file: str, parent: Path) -> typing.Any:
        f = Path(parent / Path(file)).resolve().absolute()
        if f.exists():
            items = load_json(f)
            if isinstance(items, list):
                return [Prompty.normalize(value, parent) for value in items]
            elif isinstance(items, dict):
                return {key: Prompty.normalize(value, parent) for key, value in items.items()}
            else:
                return items
        else:
            raise FileNotFoundError(f"File {file} not found")

    @staticmethod
    async def _process_file_async(file: str, parent: Path) -> typing.Any:
        f = Path(parent / Path(file)).resolve().absolute()
        if f.exists():
            items = await load_json_async(f)
            if isinstance(items, list):
                return [Prompty.normalize(value, parent) for value in items]
            elif isinstance(items, dict):
                return {key: Prompty.normalize(value, parent) for key, value in items.items()}
            else:
                return items
        else:
            raise FileNotFoundError(f"File {file} not found")

    @staticmethod
    def extract_slots(root: str, attribute: typing.Any, acc: list[dict[str, str]]) -> list[dict[str, str]]:

        if isinstance(attribute, str):
            if attribute.startswith("${env"):
                variable = attribute[2:-1].split(":")
                if len(variable) < 2:
                    raise ValueError(f"Invalid environment/slot variable {attribute}")
                if len(variable) == 2:
                    return [{"name": root, "key": variable[1]}]
                else:
                    return [{"name": root, "key": variable[1], "default": variable[2]}]
            else:
                return acc

        if isinstance(attribute, list):
            params: list[dict[str, str]] = []
            for i, v in enumerate(attribute):
                if isinstance(v, dict) and "name" in v:
                    params = [
                        *params,
                        *Prompty.extract_slots(f'{root}.{v["name"]}', v, acc),
                    ]
                elif isinstance(v, dict) and "id" in v:
                    params = [
                        *params,
                        *Prompty.extract_slots(f'{root}.{v["id"]}', v, acc),
                    ]
                else:
                    params = [*params, *Prompty.extract_slots(f"{root}.{i}", v, acc)]

            return [*acc, *params]

        if isinstance(attribute, dict):
            params = []
            for k, v in attribute.items():
                params = [*params, *Prompty.extract_slots(f"{root}.{k}", v, acc)]
            return [*acc, *params]

        return acc

    @staticmethod
    def normalize(attribute: typing.Any, parent: Path) -> typing.Any:
        if isinstance(attribute, str):
            attribute = attribute.strip()
            if attribute.startswith("${file:") and attribute.endswith("}"):
                # check if env or file
                variable = attribute[2:-1].split(":")
                return Prompty._process_file(variable[1], parent)
            else:
                return attribute
        elif isinstance(attribute, list):
            return [Prompty.normalize(value, parent) for value in attribute]
        elif isinstance(attribute, dict):
            return {key: Prompty.normalize(value, parent) for key, value in attribute.items()}
        else:
            return attribute

    @staticmethod
    async def normalize_async(attribute: typing.Any, parent: Path) -> typing.Any:
        if isinstance(attribute, str):
            attribute = attribute.strip()
            if attribute.startswith("${file") and attribute.endswith("}"):
                variable = attribute[2:-1].split(":")
                return Prompty._process_file(variable[1], parent)
            else:
                return attribute
        elif isinstance(attribute, list):
            return [await Prompty.normalize_async(value, parent) for value in attribute]
        elif isinstance(attribute, dict):
            return {key: await Prompty.normalize_async(value, parent) for key, value in attribute.items()}
        else:
            return attribute


def param_hoisting(
    top: dict[str, typing.Any],
    bottom: dict[str, typing.Any],
    top_key: Union[str, None] = None,
) -> dict[str, typing.Any]:
    if top_key:
        new_dict = {**top[top_key]} if top_key in top else {}
    else:
        new_dict = {**top}
    for key, value in bottom.items():
        if key not in new_dict:
            new_dict[key] = value
    return new_dict


class PromptyStream(Iterator):
    """PromptyStream class to iterate over LLM stream.
    Necessary for Prompty to handle streaming data when tracing."""

    def __init__(self, name: str, iterator: Iterator):
        self.name = name
        self.iterator = iterator
        self.items: list[typing.Any] = []
        self.__name__ = "PromptyStream"

    def __iter__(self):
        return self

    def __next__(self):
        try:
            # enumerate but add to list
            o = self.iterator.__next__()
            self.items.append(o)
            return o

        except StopIteration:
            # StopIteration is raised
            # contents are exhausted
            if len(self.items) > 0:
                with Tracer.start("PromptyStream") as trace:
                    trace("signature", f"{self.name}.PromptyStream")
                    trace("inputs", "None")
                    trace("result", [to_dict(s) for s in self.items])

            raise StopIteration


class AsyncPromptyStream(AsyncIterator):
    """AsyncPromptyStream class to iterate over LLM stream.
    Necessary for Prompty to handle streaming data when tracing."""

    def __init__(self, name: str, iterator: AsyncIterator):
        self.name = name
        self.iterator = iterator
        self.items: list[typing.Any] = []
        self.__name__ = "AsyncPromptyStream"

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            # enumerate but add to list
            o = await self.iterator.__anext__()
            self.items.append(o)
            return o

        except StopAsyncIteration:
            # StopIteration is raised
            # contents are exhausted
            if len(self.items) > 0:
                with Tracer.start("AsyncPromptyStream") as trace:
                    trace("signature", f"{self.name}.AsyncPromptyStream")
                    trace("inputs", "None")
                    trace("result", [to_dict(s) for s in self.items])

            raise StopAsyncIteration



import json
import sys
from pathlib import Path

if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("Invalid coverage file")
        exit(-1)

    c = Path(sys.argv[1])
    # c = Path("coverage.json")
    if not c.exists():
        print("Invalid coverage file")
        exit(-1)

    # read the coverage file
    with open(c, encoding="utf-8") as f:
        cov = json.load(f)

    print("# Coverage report")
    print("pytest-cov coverage report")
    print("")
    print("## Totals")
    # print the coverage totals
    if "totals" in cov:
        print("Item | Value")
        print("----|------")
        for k, v in cov["totals"].items():
            print(f"{k} | {v}")

    # print the coverage file
    print("\n## Files")
    if "files" in cov:
        print("| file | covered_lines | num_statements | percent_covered | missing_lines | excluded_lines |")
        print("| ---- | ------------: | -------------: | ---------------:| ------------: | ---------------: |")
        for k, v in cov["files"].items():
            pct = int(v["summary"]["percent_covered_display"])
            pct_text = str(pct) + "%"
            if pct < 96:
                pct_text = f"**{pct_text}**"

            print(f"| `{k}` |", end="")
            print(f" {v['summary']['covered_lines']} |", end="")
            print(f" {v['summary']['num_statements']} |", end="")
            print(f" {pct_text} |", end="")
            print(f" {v['summary']['missing_lines']} |", end="")
            print(f" {v['summary']['excluded_lines']} |")



import traceback
import typing
import uuid
from pathlib import Path
from typing import Union

from ._version import VERSION
from .core import (
    Prompty,
    param_hoisting,
)
from .invoker import InvokerFactory
from .parsers import PromptyChatParser
from .renderers import Jinja2Renderer, MustacheRenderer
from .tracer import trace
from .utils import (
    get_json_type,
    load_global_config,
    load_global_config_async,
    load_prompty,
    load_prompty_async,
)

__version__ = VERSION

InvokerFactory.add_renderer("jinja2", Jinja2Renderer)
InvokerFactory.add_renderer("mustache", MustacheRenderer)
InvokerFactory.add_parser("prompty.chat", PromptyChatParser)
InvokerFactory.add_parser("prompty.agent", PromptyChatParser)


def _load_with_slots(
    attributes: dict[str, typing.Any],
    content: str,
    global_config: dict[str, typing.Any],
    p: Path,
) -> Prompty:
    # load prompty dictionary from file
    prompty_dictionary = Prompty.load_manifest(attributes, content, global_config)

    slots = Prompty.extract_slots("manifest", prompty_dictionary, [])

    prompty = Prompty.load_raw(prompty_dictionary, p)
    prompty.slots = slots

    return prompty


@trace(description="Create a headless prompty object for programmatic use.")
def headless(
    api: str,
    content: Union[str, list[str], dict],
    connection: dict[str, typing.Any] = {},
    options: dict[str, typing.Any] = {},
    config: str = "default",
) -> Prompty:
    """Create a headless prompty object for programmatic use.

    Parameters
    ----------
    api : str
        The API to use for the model
    content : str | List[str] | dict
        The content to process
    configuration : Dict[str, any], optional
        The configuration to use, by default {}
    options : Dict[str, any], optional
        The options to use, by default {}
    config : str, optional
        The config to use, by default "default"

    Returns
    -------
    Prompty
        The headless prompty object

    Example
    -------
    >>> import prompty
    >>> p = prompty.headless(
            api="embedding",
            configuration={"type": "azure", "azure_deployment": "text-embedding-ada-002"},
            content="hello world",
        )
    >>> emb = prompty.execute(p)

    """

    # get caller's path (to get relative path for prompty.json)
    caller = Path(traceback.extract_stack()[-3].filename)

    attributes = {
        "template": {
            "format": "NOOP",
            "parser": "NOOP",
        },
        "model": {
            "api": api,
            "connection": connection,
            "options": options,
        },
    }

    # load global configuration
    global_config = load_global_config(caller.parent, config)
    prompty = _load_with_slots(attributes, "", global_config, caller.parent)
    prompty.content = content
    prompty.file = ""

    return prompty


@trace(description="Create a headless prompty object for programmatic use.")
async def headless_async(
    api: str,
    content: Union[str, list[str], dict],
    connection: dict[str, typing.Any] = {},
    options: dict[str, typing.Any] = {},
    config: str = "default",
) -> Prompty:
    """Create a headless prompty object for programmatic use.

    Parameters
    ----------
    api : str
        The API to use for the model
    content : str | List[str] | dict
        The content to process
    configuration : Dict[str, any], optional
        The configuration to use, by default {}
    options : Dict[str, any], optional
        The options to use, by default {}
    connection : str, optional
        The connection to use, by default "default"

    Returns
    -------
    Prompty
        The headless prompty object

    Example
    -------
    >>> import prompty
    >>> p = await prompty.headless_async(
            api="embedding",
            configuration={"type": "azure", "azure_deployment": "text-embedding-ada-002"},
            content="hello world",
        )
    >>> emb = prompty.execute(p)

    """

    # get caller's path (to get relative path for prompty.json)
    caller = Path(traceback.extract_stack()[-3].filename)

    attributes = {
        "template": {
            "format": "NOOP",
            "parser": "NOOP",
        },
        "model": {
            "api": api,
            "connection": connection,
            "options": options,
        },
    }

    # load global configuration
    global_config = await load_global_config_async(caller.parent, config)
    prompty = _load_with_slots(attributes, "", global_config, caller.parent)
    prompty.content = content
    prompty.file = ""

    return prompty


@trace(description="Load a prompty file.")
def load(prompty_file: str, config: str = "default") -> Prompty:
    """Load a prompty file.

    Parameters
    ----------
    prompty_file : str
        The path to the prompty file
    config : str, optional
        The config to use, by default "default"

    Returns
    -------
    Prompty
        The loaded prompty object

    Example
    -------
    >>> import prompty
    >>> p = prompty.load("prompts/basic.prompty")
    >>> print(p)
    """

    p = Path(prompty_file)
    if not p.is_absolute():
        # get caller's path (take into account trace frame)
        caller = Path(traceback.extract_stack()[-3].filename)
        p = Path(caller.parent / p).resolve().absolute()

    # load dictionary from prompty file
    matter = load_prompty(p)
    atttributes = matter.pop("attributes", {})
    # expand ${file} includes
    attributes = Prompty.normalize(atttributes, p.parent)

    # load global configuration
    global_config = load_global_config(p.parent, config)

    return _load_with_slots(attributes, matter["body"], global_config, p)


@trace(description="Load a prompty file.")
async def load_async(prompty_file: str, config: str = "default") -> Prompty:
    """Load a prompty file.

    Parameters
    ----------
    prompty_file : str
        The path to the prompty file
    config : str, optional
        The config to use, by default "default"

    Returns
    -------
    Prompty
        The loaded prompty object

    Example
    -------
    >>> import prompty
    >>> p = prompty.load("prompts/basic.prompty")
    >>> print(p)
    """

    p = Path(prompty_file)
    if not p.is_absolute():
        # get caller's path (take into account trace frame)
        caller = Path(traceback.extract_stack()[-3].filename)
        p = Path(caller.parent / p).resolve().absolute()

    # load dictionary from prompty file
    matter = await load_prompty_async(p)
    atttributes = matter.pop("attributes", {})
    # expand ${file} includes
    attributes = await Prompty.normalize_async(atttributes, p.parent)

    # load global configuration
    global_config = await load_global_config_async(p.parent, config)

    return _load_with_slots(attributes, matter["body"], global_config, p)


def _validate_inputs(prompt: Prompty, inputs: dict[str, typing.Any], merge_sample: bool = False):
    if merge_sample:
        inputs = param_hoisting(inputs, prompt.get_sample())

    clean_inputs = {}
    for input in prompt.inputs:
        # thread managed seperately
        if input.name == "thread":
            continue

        if input.name in inputs:
            if input.type != get_json_type(type(inputs[input.name])):
                raise ValueError(
                    f"Type mismatch for input property {input.name}: input type ({inputs[input.name].type}) != sample type ({input.type})"
                )
            clean_inputs[input.name] = inputs[input.name]
        else:
            if input.default is not None:
                clean_inputs[input.name] = input.default
            else:
                raise ValueError(f"Missing input property {input.name}")

    # check stra inputs
    invalid: list[str] = []
    for k, v in inputs.items():
        if prompt.get_input(k) is None:
            invalid.append(k)

    if len(invalid) > 0:
        raise ValueError(f"The following are not valid inputs: [{','.join(invalid)}]")

    return clean_inputs


@trace(description="Prepare the inputs for the prompt.")
def prepare(
    prompt: Prompty,
    inputs: dict[str, typing.Any] = {},
    merge_sample: bool = False,
):
    """Prepare the inputs for the prompt.

    Parameters
    ----------
    prompt : Prompty
        The prompty object
    inputs : Dict[str, any], optional
        The inputs to the prompt, by default {}

    Returns
    -------
    dict
        The prepared and hidrated template shaped to the LLM model

    Example
    -------
    >>> import prompty
    >>> p = prompty.load("prompts/basic.prompty")
    >>> inputs = {"name": "John Doe"}
    >>> content = prompty.prepare(p, inputs)
    """
    values = _validate_inputs(prompt, inputs, merge_sample)

    # add nonce for this run
    prompt.template.nonce = uuid.uuid4().hex

    render = InvokerFactory.run_renderer(prompt, values, prompt.content)
    result = InvokerFactory.run_parser(prompt, render)

    return result


@trace(description="Prepare the inputs for the prompt.")
async def prepare_async(
    prompt: Prompty,
    inputs: dict[str, typing.Any] = {},
    merge_sample: bool = False,
):
    """Prepare the inputs for the prompt.

    Parameters
    ----------
    prompt : Prompty
        The prompty object
    inputs : Dict[str, any], optional
        The inputs to the prompt, by default {}

    Returns
    -------
    dict
        The prepared and hidrated template shaped to the LLM model

    Example
    -------
    >>> import prompty
    >>> p = prompty.load("prompts/basic.prompty")
    >>> inputs = {"name": "John Doe"}
    >>> content = await prompty.prepare_async(p, inputs)
    """
    values = _validate_inputs(prompt, inputs, merge_sample)

    # add nonce for this run
    prompt.template.nonce = uuid.uuid4().hex

    render = await InvokerFactory.run_renderer_async(prompt, values, prompt.content)
    result = await InvokerFactory.run_parser_async(prompt, render)

    return result


@trace(description="Run the prepared Prompty content against the model.")
def run(
    prompt: Prompty,
    content: Union[dict, list, str],
    connection: dict[str, typing.Any] = {},
    options: dict[str, typing.Any] = {},
    slots: dict[str, typing.Any] = {},
    raw: bool = False,
):
    """Run the prepared Prompty content.

    Parameters
    ----------
    prompt : Prompty
        The prompty object
    content : dict | list | str
        The content to process
    connection : Dict[str, any], optional
        The connection to use, by default {}
    options : Dict[str, any], optional
        The options to use, by default {}
    raw : bool, optional
        Whether to skip processing, by default False

    Returns
    -------
    any
        The result of the prompt

    Example
    -------
    >>> import prompty
    >>> p = prompty.load("prompts/basic.prompty")
    >>> inputs = {"name": "John Doe"}
    >>> content = prompty.prepare(p, inputs)
    >>> result = prompty.run(p, content)
    """

    if connection != {}:
        prompt.model.connection = param_hoisting(connection, prompt.model.connection)

    if options != {}:
        prompt.model.options = param_hoisting(options, prompt.model.options)

    # map slots (if any keys are in the slots)
    if slots != {}:
        for item in prompt.slots:
            if item["key"] in slots:
                item["value"] = slots[item["key"]]

    result = InvokerFactory.run_executor(prompt, content)
    if not raw:
        result = InvokerFactory.run_processor(prompt, result)

    return result


@trace(description="Run the prepared Prompty content against the model.")
async def run_async(
    prompt: Prompty,
    content: Union[dict, list, str],
    connection: dict[str, typing.Any] = {},
    options: dict[str, typing.Any] = {},
    slots: dict[str, typing.Any] = {},
    raw: bool = False,
):
    """Run the prepared Prompty content.

    Parameters
    ----------
    prompt : Prompty
        The prompty object
    content : dict | list | str
        The content to process
    connection : Dict[str, any], optional
        The connection to use, by default {}
    options : Dict[str, any], optional
        The options to use, by default {}
    raw : bool, optional
        Whether to skip processing, by default False

    Returns
    -------
    any
        The result of the prompt

    Example
    -------
    >>> import prompty
    >>> p = prompty.load("prompts/basic.prompty")
    >>> inputs = {"name": "John Doe"}
    >>> content = await prompty.prepare_async(p, inputs)
    >>> result = await prompty.run_async(p, content)
    """

    if connection != {}:
        prompt.model.connection = param_hoisting(connection, prompt.model.connection)

    if options != {}:
        prompt.model.options = param_hoisting(options, prompt.model.options)

    # map slots (if any keys are in the slots)
    if slots != {}:
        for item in prompt.slots:
            if item["key"] in slots:
                item["value"] = slots[item["key"]]

    result = await InvokerFactory.run_executor_async(prompt, content)
    if not raw:
        result = await InvokerFactory.run_processor_async(prompt, result)

    return result


@trace(description="Execute a prompty")
def execute(
    prompt: Union[str, Prompty],
    connection: dict[str, typing.Any] = {},
    options: dict[str, typing.Any] = {},
    inputs: dict[str, typing.Any] = {},
    env: dict[str, typing.Any] = {},
    raw: bool = False,
    merge_sample: bool = False,
    config_name: str = "default",
):
    """Execute a prompty.

    Parameters
    ----------
    prompt : Union[str, Prompty]
        The prompty object or path to the prompty file
    connection : Dict[str, any], optional
        The connection to use, by default {}
    options : Dict[str, any], optional
        The options to use, by default {}
    inputs : Dict[str, any], optional
        The inputs to the prompt, by default {}
    raw : bool, optional
        Whether to skip processing, by default False
    connection : str, optional
        The connection to use, by default "default"

    Returns
    -------
    any
        The result of the prompt

    Example
    -------
    >>> import prompty
    >>> inputs = {"name": "John Doe"}
    >>> result = prompty.execute("prompts/basic.prompty", inputs=inputs)
    """
    if isinstance(prompt, (str, Path)):
        path = Path(prompt)
        if not path.is_absolute():
            # get caller's path (take into account trace frame)
            caller = Path(traceback.extract_stack()[-3].filename)
            path = Path(caller.parent / path).resolve().absolute()
        prompt = load(path, config_name)

    # prepare content
    content = prepare(prompt, inputs, merge_sample)

    # run LLM model
    result = run(prompt, content, connection, options, env, raw)

    return result


@trace(description="Execute a prompty")
async def execute_async(
    prompt: Union[str, Prompty],
    connection: dict[str, typing.Any] = {},
    options: dict[str, typing.Any] = {},
    inputs: dict[str, typing.Any] = {},
    env: dict[str, typing.Any] = {},
    raw: bool = False,
    merge_sample: bool = False,
    config_name: str = "default",
):
    """Execute a prompty.

    Parameters
    ----------
    prompt : Union[str, Prompty]
        The prompty object or path to the prompty file
    connection : Dict[str, any], optional
        The connection to use, by default {}
    options : Dict[str, any], optional
        The options to use, by default {}
    inputs : Dict[str, any], optional
        The inputs to the prompt, by default {}
    raw : bool, optional
        Whether to skip processing, by default False
    connection : str, optional
        The connection to use, by default "default"

    Returns
    -------
    any
        The result of the prompt

    Example
    -------
    >>> import prompty
    >>> inputs = {"name": "John Doe"}
    >>> result = await prompty.execute_async("prompts/basic.prompty", inputs=inputs)
    """
    if isinstance(prompt, (str, Path)):
        path = Path(prompt)
        if not path.is_absolute():
            # get caller's path (take into account trace frame)
            caller = Path(traceback.extract_stack()[-3].filename)
            path = Path(caller.parent / path).resolve().absolute()
        prompt = await load_async(path, config_name)

    # prepare content
    content = await prepare_async(prompt, inputs, merge_sample)

    # run LLM model
    result = await run_async(prompt, content, connection, options, env, raw)

    return result



import abc
import os
import typing
from typing import Callable, Literal, Union

from .core import Prompty
from .tracer import trace


class Invoker(abc.ABC):
    """Abstract class for Invoker

    Attributes
    ----------
    prompty : Prompty
        The prompty object
    name : str
        The name of the invoker

    """

    def __init__(self, prompty: Prompty) -> None:
        self.prompty = prompty
        self.name = self.__class__.__name__
        self._resolved = False

    @abc.abstractmethod
    def invoke(self, data: typing.Any) -> typing.Any:
        """Abstract method to invoke the invoker

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """
        pass

    @abc.abstractmethod
    async def invoke_async(self, data: typing.Any) -> typing.Any:
        """Abstract method to invoke the invoker asynchronously

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """
        pass

    @trace
    def run(self, data: typing.Any) -> typing.Any:
        """Method to run the invoker

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """

        return self.invoke(data)

    @trace
    async def run_async(self, data: typing.Any) -> typing.Any:
        """Method to run the invoker asynchronously

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """
        return await self.invoke_async(data)

    @staticmethod
    def _process_env(variable: str, env_error=True, default: Union[str, None] = None) -> typing.Any:
        if variable in os.environ.keys():
            return os.environ[variable]
        else:
            if default:
                return default
            if env_error:
                raise ValueError(f"Variable {variable} not found in environment")

            return ""

    def resolve_model(self) -> None:
        """Resolve model variables"""

        # only resolve once
        if self._resolved:
            return

        self.prompty.model.connection = self.resolve_slots("manifest.model.connection", self.prompty.model.connection)

        self.prompty.model.options = self.resolve_slots("manifest.model.options", self.prompty.model.options)
        self._resolved = True

    def resolve_slots(self, root: str, attribute: typing.Any) -> typing.Any:
        if isinstance(attribute, str):
            if attribute.startswith("${env"):
                variable = attribute[2:-1].split(":")
                if len(variable) < 2:
                    raise ValueError(f"Invalid environment/slot variable {attribute}")

                key = variable[1].strip()
                idx = next(
                    (
                        index
                        for index, item in enumerate(self.prompty.slots)
                        if item["key"] == key and item["name"].endswith(root)
                    ),
                    -1,
                )

                if idx == -1:
                    raise ValueError(f"Slot {key} not found in Prompty slots for {root}")
                else:
                    # use value if it exists
                    if "value" in self.prompty.slots[idx]:
                        return self.prompty.slots[idx]["value"]
                    # otherwise, use env variable
                    else:
                        # cache the env variable in the slot
                        # if the env variable is not found, use the default value
                        self.prompty.slots[idx]["value"] = self._process_env(
                            key, env_error=True, default=self.prompty.slots[idx].get("default", None)
                        )
                        return self.prompty.slots[idx]["value"]
            else:
                return attribute

        if isinstance(attribute, list):

            def get_key(v: dict[str, str], i) -> str:
                if "name" in v:
                    return v["name"]
                elif "id" in v:
                    return v["id"]
                else:
                    return str(i)

            return [self.resolve_slots(f"{root}.{get_key(v, i)}", v) for i, v in enumerate(attribute)]

        if isinstance(attribute, dict):
            return {key: self.resolve_slots(f"{root}.{key}", value) for key, value in attribute.items()}

        return attribute


class Renderer(Invoker):
    """Abstract class for Renderer

    Attributes
    ----------
    prompty : Prompty
        The prompty object
    name : str
        The name of the renderer

    """

    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)

    def run(self, data: typing.Any) -> typing.Any:
        """Method to run the invoker

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """

        # check if parser inherits from Parser
        parser = InvokerFactory._get_invoker("parser", self.prompty)
        if isinstance(parser, Parser):
            self.prompty.template.content = parser.sanitize(self.prompty.content)

        return self.invoke(data)

    async def run_async(self, data: typing.Any) -> typing.Any:
        """Method to run the invoker asynchronously

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """

        # check if parser inherits from Parser
        parser = InvokerFactory._get_invoker("parser", self.prompty)
        if isinstance(parser, Parser):
            self.prompty.template.content = parser.sanitize(self.prompty.content)

        return await self.invoke_async(data)


class Parser(Invoker):
    """Abstract class for Parser

    Attributes
    ----------
    prompty : Prompty
        The prompty object
    name : str
        The name of the parser

    """

    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)

    @abc.abstractmethod
    def sanitize(self, data: typing.Any) -> typing.Any:
        """Abstract method to sanitize template

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """
        pass

    @abc.abstractmethod
    def process(self, data: typing.Any) -> typing.Any:
        """Method to process parsed content

        Parameters
        ----------
        data : any
            The parsed content

        Returns
        -------
        any
            Processed parsed data
        """
        pass

    def run(self, data: typing.Any) -> typing.Any:
        """Method to run the invoker

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """

        parsed = self.invoke(data)
        return self.process(parsed)

    async def run_async(self, data: typing.Any) -> typing.Any:
        """Method to run the invoker asynchronously

        Parameters
        ----------
        data : any
            The data to be invoked

        Returns
        -------
        any
            The invoked
        """
        parsed = await self.invoke_async(data)
        return self.process(parsed)


class InvokerFactory:
    """Factory class for Invoker"""

    _renderers: dict[str, type[Invoker]] = {}
    _parsers: dict[str, type[Invoker]] = {}
    _executors: dict[str, type[Invoker]] = {}
    _processors: dict[str, type[Invoker]] = {}

    @classmethod
    def add_renderer(cls, name: str, invoker: type[Invoker]) -> None:
        cls._renderers[name] = invoker

    @classmethod
    def add_parser(cls, name: str, invoker: type[Invoker]) -> None:
        cls._parsers[name] = invoker

    @classmethod
    def add_executor(cls, name: str, invoker: type[Invoker]) -> None:
        cls._executors[name] = invoker

    @classmethod
    def add_processor(cls, name: str, invoker: type[Invoker]) -> None:
        cls._processors[name] = invoker

    @classmethod
    def register_renderer(cls, name: str) -> Callable:

        def inner_wrapper(wrapped_class: type[Invoker]) -> type[Invoker]:
            cls._renderers[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def register_parser(cls, name: str) -> Callable:

        def inner_wrapper(wrapped_class: type[Invoker]) -> type[Invoker]:
            cls._parsers[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def register_executor(cls, name: str) -> Callable:

        def inner_wrapper(wrapped_class: type[Invoker]) -> type[Invoker]:
            cls._executors[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def register_processor(cls, name: str) -> Callable:

        def inner_wrapper(wrapped_class: type[Invoker]) -> type[Invoker]:
            cls._processors[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def _get_name(
        cls,
        type: Literal["renderer", "parser", "executor", "processor"],
        prompty: Prompty,
    ) -> str:
        if type == "renderer":
            return prompty.template.format
        elif type == "parser":
            return f"{prompty.template.parser}.{prompty.model.api}"
        elif type == "executor":
            return prompty.model.connection["type"]
        elif type == "processor":
            return prompty.model.connection["type"]
        else:
            raise ValueError(f"Type {type} not found")

    @classmethod
    def _get_invoker(
        cls,
        type: Literal["renderer", "parser", "executor", "processor"],
        prompty: Prompty,
    ) -> Invoker:
        if type == "renderer":
            name = prompty.template.format
            if name not in cls._renderers:
                raise ValueError(f"Renderer {name} not found")

            return cls._renderers[name](prompty)

        elif type == "parser":
            name = f"{prompty.template.parser}.{prompty.model.api}"
            if name not in cls._parsers:
                raise ValueError(f"Parser {name} not found")

            return cls._parsers[name](prompty)

        elif type == "executor":
            name = prompty.model.connection["type"]
            if name not in cls._executors:
                raise ValueError(f"Executor {name} not found")

            return cls._executors[name](prompty)

        elif type == "processor":
            name = prompty.model.connection["type"]
            if name not in cls._processors:
                raise ValueError(f"Processor {name} not found")

            return cls._processors[name](prompty)

        else:
            raise ValueError(f"Type {type} not found")

    @classmethod
    def run(
        cls,
        type: Literal["renderer", "parser", "executor", "processor"],
        prompty: Prompty,
        data: typing.Any,
        default: typing.Any = None,
    ):
        name = cls._get_name(type, prompty)
        if name.startswith("NOOP") and default is not None:
            return default
        elif name.startswith("NOOP"):
            return data

        invoker = cls._get_invoker(type, prompty)
        value = invoker.run(data)
        return value

    @classmethod
    async def run_async(
        cls,
        type: Literal["renderer", "parser", "executor", "processor"],
        prompty: Prompty,
        data: typing.Any,
        default: typing.Any = None,
    ):
        name = cls._get_name(type, prompty)
        if name.startswith("NOOP") and default is not None:
            return default
        elif name.startswith("NOOP"):
            return data
        invoker = cls._get_invoker(type, prompty)
        value = await invoker.run_async(data)
        return value

    @classmethod
    def run_renderer(cls, prompty: Prompty, data: typing.Any, default: typing.Any = None) -> typing.Any:
        return cls.run("renderer", prompty, data, default)

    @classmethod
    async def run_renderer_async(cls, prompty: Prompty, data: typing.Any, default: typing.Any = None) -> typing.Any:
        return await cls.run_async("renderer", prompty, data, default)

    @classmethod
    def run_parser(cls, prompty: Prompty, data: typing.Any, default: typing.Any = None) -> typing.Any:
        return cls.run("parser", prompty, data, default)

    @classmethod
    async def run_parser_async(cls, prompty: Prompty, data: typing.Any, default: typing.Any = None) -> typing.Any:
        return await cls.run_async("parser", prompty, data, default)

    @classmethod
    def run_executor(cls, prompty: Prompty, data: typing.Any, default: typing.Any = None) -> typing.Any:
        return cls.run("executor", prompty, data, default)

    @classmethod
    async def run_executor_async(cls, prompty: Prompty, data: typing.Any, default: typing.Any = None) -> typing.Any:
        return await cls.run_async("executor", prompty, data, default)

    @classmethod
    def run_processor(cls, prompty: Prompty, data: typing.Any, default: typing.Any = None) -> typing.Any:
        return cls.run("processor", prompty, data, default)

    @classmethod
    async def run_processor_async(cls, prompty: Prompty, data: typing.Any, default: typing.Any = None) -> typing.Any:
        return await cls.run_async("processor", prompty, data, default)


class InvokerException(Exception):
    """Exception class for Invoker"""

    def __init__(self, message: str, type: str) -> None:
        super().__init__(message)
        self.type = type

    def __str__(self) -> str:
        return f"{super().__str__()}. Make sure to pip install any necessary package extras (i.e. could be something like `pip install prompty[{self.type}]`) for {self.type} as well as import the appropriate invokers (i.e. could be something like `import prompty.{self.type}`)."


@InvokerFactory.register_renderer("NOOP")
@InvokerFactory.register_parser("NOOP")
@InvokerFactory.register_executor("NOOP")
@InvokerFactory.register_processor("NOOP")
@InvokerFactory.register_parser("prompty.embedding")
@InvokerFactory.register_parser("prompty.image")
@InvokerFactory.register_parser("prompty.completion")
class NoOp(Invoker):
    def invoke(self, data: typing.Any) -> typing.Any:
        return data

    async def invoke_async(self, data: str) -> str:
        return self.invoke(data)



import typing
from collections.abc import Iterator

from openai import OpenAI

from prompty.tracer import Tracer

from .._version import VERSION
from ..core import Prompty, PromptyStream
from ..invoker import Invoker, InvokerFactory


@InvokerFactory.register_executor("openai")
class OpenAIExecutor(Invoker):
    """OpenAI Executor"""

    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)
        self.kwargs = {
            key: value for key, value in self.prompty.model.connection.items() if key != "type" and key != "name"
        }

        self.api = self.prompty.model.api
        self.options = self.prompty.model.options
        self.model = self.prompty.model.connection["name"]
        self.deployment = self.prompty.model.connection["deployment"]

    def _sanitize_messages(self, data: typing.Any) -> list[dict[str, str]]:
        messages = data if isinstance(data, list) else [data]

        if self.prompty.template.strict:
            if not all([msg["nonce"] == self.prompty.template.nonce for msg in messages]):
                raise ValueError("Nonce mismatch in messages array (strict mode)")

        messages = [
            {
                **{"role": msg["role"], "content": msg["content"]},
                **({"name": msg["name"]} if "name" in msg else {}),
            }
            for msg in messages
        ]

        return messages

    def invoke(self, data: typing.Any) -> typing.Any:
        """Invoke the OpenAI API

        Parameters
        ----------
        data : any
            The data to send to the OpenAI API

        Returns
        -------
        any
            The response from the OpenAI API
        """
        with Tracer.start("OpenAI") as trace:
            trace("type", "LLM")
            trace("signature", "OpenAI.ctor")
            trace("description", "OpenAI Constructor")
            trace("inputs", self.kwargs)
            client = OpenAI(
                default_headers={
                    "User-Agent": f"prompty/{VERSION}",
                    "x-ms-useragent": f"prompty/{VERSION}",
                },
                **self.kwargs,
            )
            trace("result", client)

        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "OpenAI Prompty Execution Invoker")

            if self.api == "chat":
                trace("signature", "OpenAI.chat.completions.create")
                args = {
                    "model": self.model,
                    "messages": self._sanitize_messages(data),
                    **self.options,
                }
                trace("inputs", args)
                response = client.chat.completions.create(**args)

            elif self.api == "completion":
                trace("signature", "OpenAI.completions.create")
                args = {
                    "prompt": data.item,
                    "model": self.deployment,
                    **self.options,
                }
                trace("inputs", args)
                response = client.completions.create(**args)

            elif self.api == "embedding":
                trace("signature", "OpenAI.embeddings.create")
                args = {
                    "input": data if isinstance(data, list) else [data],
                    "model": self.deployment,
                    **self.options,
                }
                trace("inputs", args)
                response = client.embeddings.create(**args)

            elif self.api == "image":
                raise NotImplementedError("OpenAI Image API is not implemented yet")

            # stream response
            if isinstance(response, Iterator):
                stream = PromptyStream("AzureOpenAIExecutor", response)
                trace("result", stream)
                return stream
            else:
                trace("result", response)
                return response

    async def invoke_async(self, data: str) -> str:
        """Invoke the Prompty Chat Parser (Async)

        Parameters
        ----------
        data : str
            The data to parse

        Returns
        -------
        str
            The parsed data
        """
        return self.invoke(data)



from pathlib import Path

import prompty
from prompty.tracer import to_dict


class TestCore:
    def test_prompty_to_dict(self, **kwargs):
        prompt_file = "prompts/basic.prompty"
        p = prompty.load(prompt_file)
        d = to_dict(p)
        assert d["name"] == "Basic Prompt"
        assert d["model"]["connection"]["type"] == "azure"
        assert d["model"]["connection"]["azure_deployment"] == "gpt-35-turbo"
        assert d["template"]["format"] == "jinja2"

    def test_prompty_to_safe_dict(self, **kwargs):
        prompt_file_base = "prompts/fake.prompty"
        p_base = prompty.load(prompt_file_base)
        prompt_file = "prompts/chat.prompty"
        p = prompty.load(prompt_file)
        p.basePrompty = p_base
        p.inputs = {"key1": "value1", "key2": "value2"}
        p.outputs = {"key3": "value3", "key4": "value4"}
        p.file = "/path/to/file"
        d = p.to_safe_dict()
        assert d["name"] == "Basic Prompt"
        assert d["model"]["connection"]["type"] == "azure"
        assert d["model"]["connection"]["azure_deployment"] == "gpt-35-turbo"
        assert d["template"]["format"] == "jinja2"
        assert d["file"] == "/path/to/file"
        assert "basePrompty" not in d

    def test_prompty_to_safe_dict_file_path(self, **kwargs):
        prompt_file = "prompts/chat.prompty"
        p = prompty.load(prompt_file)
        p.file = Path("/path/to/file")
        d = p.to_safe_dict()
        assert d["file"] == "/path/to/file"



import prompty
from prompty.core import Prompty
from prompty.parsers import PromptyChatParser

roles = ["assistant", "function", "system", "user"]


def test_parse_with_args():
    content = 'system[key="value 1", post=false, great=True, other=3.2, pre = 2]:\nYou are an AI assistant\n who helps people find information.\nAs the assistant, you answer questions briefly, succinctly.\n\nuser:\nWhat is the meaning of life?'
    parser = PromptyChatParser(Prompty())
    messages = parser.invoke(content)
    assert len(messages) == 2
    assert messages[0]["role"] == "system"
    assert messages[0]["key"] == "value 1"
    assert messages[0]["post"] is False
    assert messages[0]["great"] is True
    assert messages[0]["other"] == 3.2
    assert messages[0]["pre"] == 2
    assert messages[1]["role"] == "user"


def test_parse_invalid_args():
    content = 'system[role="value 1", content="overwrite content",post=false, great=True, other=3.2, pre = 2]:\nYou are an AI assistant\n who helps people find information.\nAs the assistant, you answer questions briefly, succinctly.\n\nuser:\nWhat is the meaning of life?'
    parser = PromptyChatParser(Prompty())
    messages = parser.invoke(content)
    assert len(messages) == 2
    assert messages[0]["role"] == "system"
    assert (
        messages[0]["content"]
        == "You are an AI assistant\n who helps people find information.\nAs the assistant, you answer questions briefly, succinctly.\n"
    )
    assert messages[1]["role"] == "user"


def test_thread_parse():
    p = prompty.load("tools/basic.prompty")
    content = prompty.prepare(p, merge_sample=True)
    assert len(content) == 3
    assert content[0]["role"] == "system"
    assert content[1]["role"] == "thread"
    assert content[2]["role"] == "system"



import pytest
import yaml

from prompty.common import convert_function_tools
from prompty.core import Prompty

function1 = [
    {
        "type": "function",
        "function": {
            "name": "get_current_time",
            "description": "Get the current time in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city name, e.g. San Francisco",
                    },
                },
                "required": ["location"],
            },
        },
    }
]

yaml1 = """
tools:
  - id: get_current_time
    type: function
    description: Get the current time in a given location
    parameters:
      location:
        type: string
        description: The city name, e.g. San Francisco
        required: true
"""

function2 = [
    {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"},
                    "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                },
                "required": ["location"],
            },
        },
    }
]

yaml2 = """
tools:
  - id: get_current_weather
    type: function
    description: Get the current weather in a given location
    parameters:
      location:
        type: string
        description: The city and state, e.g. San Francisco, CA
        required: true
      unit:
        type: string
        enum: 
            - celsius
            - fahrenheit
"""

# combine tools
function3 = [
    {
        "type": "function",
        "function": {
            "name": "get_current_time",
            "description": "Get the current time in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city name, e.g. San Francisco",
                    },
                },
                "required": ["location"],
            },
        },
    },
    {
        "type": "function",
        "function": {
            "name": "get_current_weather",
            "description": "Get the current weather in a given location",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {"type": "string", "description": "The city and state, e.g. San Francisco, CA"},
                    "unit": {"type": "string", "enum": ["celsius", "fahrenheit"]},
                },
                "required": ["location"],
            },
        },
    },
]

yaml3 = """
tools:
  - id: get_current_time
    type: function
    description: Get the current time in a given location
    parameters:
      location:
        type: string
        description: The city name, e.g. San Francisco
        required: true
  - id: get_current_weather
    type: function
    description: Get the current weather in a given location
    parameters:
      location:
        type: string
        description: The city and state, e.g. San Francisco, CA
        required: true
      unit:
        type: string
        enum: 
            - celsius
            - fahrenheit
"""


@pytest.mark.parametrize(
    "definition,expected",
    [
        pytest.param(yaml1, function1, id="simple"),
        pytest.param(yaml2, function2, id="enum"),
        pytest.param(yaml3, function3, id="list"),
    ],
)
def test_tool_from_yaml(definition, expected):
    # test conversion from yaml to function
    fmt = yaml.load(definition, Loader=yaml.FullLoader)
    tools = Prompty.load_tools(fmt["tools"])
    obj = convert_function_tools(tools)
    assert obj == expected



import contextlib
import inspect
import json
import os
import traceback
from collections.abc import Iterator
from dataclasses import asdict, is_dataclass
from datetime import datetime
from functools import partial, wraps
from numbers import Number
from pathlib import Path
from typing import Any, Callable, Union

from ._version import VERSION


# clean up key value pairs for sensitive values
def sanitize(key: str, value: Any) -> Any:
    if isinstance(value, str) and any([s in key.lower() for s in ["key", "secret", "password", "credential"]]):
        return 10 * "*"
    elif isinstance(value, dict):
        return {k: sanitize(k, v) for k, v in value.items()}
    else:
        return value


class Tracer:
    _tracers: dict[
        str,
        Callable[[str], contextlib._GeneratorContextManager[Callable[[str, Any], None]]],
    ] = {}

    SIGNATURE = "signature"
    INPUTS = "inputs"
    RESULT = "result"

    @classmethod
    def add(
        cls,
        name: str,
        tracer: Callable[[str], contextlib._GeneratorContextManager[Callable[[str, Any], None]]],
    ) -> None:
        cls._tracers[name] = tracer

    @classmethod
    def clear(cls) -> None:
        cls._tracers = {}

    @classmethod
    @contextlib.contextmanager
    def start(
        cls, name: str, attributes: Union[dict[str, Any], None] = None
    ) -> Iterator[Callable[[str, Any], list[None]]]:
        with contextlib.ExitStack() as stack:
            traces: list[Callable[[str, Any], None]] = [
                stack.enter_context(tracer(name)) for tracer in cls._tracers.values()
            ]

            if attributes:
                for trace in traces:
                    for key, value in attributes.items():
                        trace(key, value)

            yield lambda key, value: [
                # normalize and sanitize any trace values
                trace(key, sanitize(key, to_dict(value)))
                for trace in traces
            ]


def to_dict(obj: Any) -> Any:
    # simple json types
    if isinstance(obj, str) or isinstance(obj, Number) or isinstance(obj, bool):
        return obj
    # datetime
    elif isinstance(obj, datetime):
        return obj.isoformat()
    # safe Prompty obj serialization
    elif type(obj).__name__ == "Prompty":
        obj_dict = asdict(obj)
        if "model" in obj_dict and "configuration" in obj_dict["model"]:
            obj_dict["model"]["configuration"] = sanitize("configuration", obj_dict["model"]["configuration"])
        return obj_dict
    # safe PromptyStream obj serialization
    elif type(obj).__name__ == "PromptyStream":
        return "PromptyStream"
    elif is_dataclass(obj) and not isinstance(obj, type):
        return asdict(obj)
    elif type(obj).__name__ == "AsyncPromptyStream":
        return "AsyncPromptyStream"
    # recursive list and dict
    elif isinstance(obj, list):
        return [to_dict(item) for item in obj]
    elif isinstance(obj, dict):
        return {k: v if isinstance(v, str) else to_dict(v) for k, v in obj.items()}
    elif isinstance(obj, Path):
        return str(obj)
    # cast to string otherwise...
    else:
        return str(obj)


def _name(func: Callable, args):
    if hasattr(func, "__qualname__"):
        signature = f"{func.__module__}.{func.__qualname__}"
    else:
        signature = f"{func.__module__}.{func.__name__}"

    # core invoker gets special treatment prompty.invoker.Invoker
    core_invoker = signature.startswith("prompty.invoker.Invoker.run")
    if core_invoker:
        name = type(args[0]).__name__
        if signature.endswith("async"):
            signature = f"{args[0].__module__}.{args[0].__class__.__name__}.invoke_async"
        else:
            signature = f"{args[0].__module__}.{args[0].__class__.__name__}.invoke"
    else:
        name = func.__name__

    return name, signature


def _inputs(func: Callable, args, kwargs) -> dict:
    ba = inspect.signature(func).bind(*args, **kwargs)
    ba.apply_defaults()

    inputs = {k: to_dict(v) for k, v in ba.arguments.items() if k != "self"}

    return inputs


def _results(result: Any) -> Any:
    return to_dict(result) if result is not None else "None"


def _trace_sync(func: Callable, **okwargs: Any) -> Callable:

    @wraps(func)
    def wrapper(*args, **kwargs):
        name, signature = _name(func, args)
        altname: Union[str, None] = None
        # special case
        if "name" in okwargs:
            altname = name
            name = okwargs["name"]
            del okwargs["name"]

        with Tracer.start(name) as trace:
            if altname is not None:
                trace("function", altname)

            trace("signature", signature)

            # support arbitrary keyword
            # arguments for trace decorator
            for k, v in okwargs.items():
                trace(k, to_dict(v))

            inputs = _inputs(func, args, kwargs)
            trace("inputs", inputs)

            try:
                result = func(*args, **kwargs)
                trace("result", _results(result))
            except Exception as e:
                trace(
                    "result",
                    {
                        "exception": {
                            "type": type(e),
                            "traceback": (traceback.format_tb(tb=e.__traceback__) if e.__traceback__ else None),
                            "message": str(e),
                            "args": to_dict(e.args),
                        }
                    },
                )
                raise e

            return result

    return wrapper


def _trace_async(func: Callable, **okwargs: Any) -> Callable:

    @wraps(func)
    async def wrapper(*args, **kwargs):
        name, signature = _name(func, args)
        altname: Union[str, None] = None
        # special case
        if "name" in okwargs:
            altname = name
            name = okwargs["name"]
            del okwargs["name"]

        with Tracer.start(name) as trace:
            if altname is not None:
                trace("function", altname)

            trace("signature", signature)

            # support arbitrary keyword
            # arguments for trace decorator
            for k, v in okwargs.items():
                trace(k, to_dict(v))

            inputs = _inputs(func, args, kwargs)
            trace("inputs", inputs)
            try:
                result = await func(*args, **kwargs)
                trace("result", _results(result))
            except Exception as e:
                trace(
                    "result",
                    {
                        "exception": {
                            "type": type(e),
                            "traceback": (traceback.format_tb(tb=e.__traceback__) if e.__traceback__ else None),
                            "message": str(e),
                            "args": to_dict(e.args),
                        }
                    },
                )
                raise e

            return result

    return wrapper


def trace(func: Union[Callable, None] = None, **kwargs: Any) -> Callable:
    if func is None:
        return partial(trace, **kwargs)
    wrapped_method = _trace_async if inspect.iscoroutinefunction(func) else _trace_sync
    return wrapped_method(func, **kwargs)


class PromptyTracer:
    def __init__(self, output_dir: Union[str, None] = None) -> None:
        if output_dir:
            self.output = Path(output_dir).resolve().absolute()
        else:
            self.output = Path(Path(os.getcwd()) / ".runs").resolve().absolute()

        if not self.output.exists():
            self.output.mkdir(parents=True, exist_ok=True)

        self.stack: list[dict[str, Any]] = []

    @contextlib.contextmanager
    def tracer(self, name: str) -> Iterator[Callable[[str, Any], None]]:
        try:
            self.stack.append({"name": name})
            frame = self.stack[-1]
            frame["__time"] = {
                "start": datetime.now(),
            }

            def add(key: str, value: Any) -> None:
                if key not in frame:
                    frame[key] = value
                # multiple values creates list
                else:
                    if isinstance(frame[key], list):
                        frame[key].append(value)
                    else:
                        frame[key] = [frame[key], value]

            yield add
        finally:
            frame = self.stack.pop()
            start: datetime = frame["__time"]["start"]
            end: datetime = datetime.now()

            # add duration to frame
            frame["__time"] = {
                "start": start.strftime("%Y-%m-%dT%H:%M:%S.%f"),
                "end": end.strftime("%Y-%m-%dT%H:%M:%S.%f"),
                "duration": int((end - start).total_seconds() * 1000),
            }

            # hoist usage to parent frame
            if "result" in frame and isinstance(frame["result"], dict):
                if "usage" in frame["result"]:
                    frame["__usage"] = self.hoist_item(
                        frame["result"]["usage"],
                        frame["__usage"] if "__usage" in frame else {},
                    )

            # streamed results may have usage as well
            if "result" in frame and isinstance(frame["result"], list):
                for result in frame["result"]:
                    if isinstance(result, dict) and "usage" in result and isinstance(result["usage"], dict):
                        frame["__usage"] = self.hoist_item(
                            result["usage"],
                            frame["__usage"] if "__usage" in frame else {},
                        )

            # add any usage frames from below
            if "__frames" in frame:
                for child in frame["__frames"]:
                    if "__usage" in child:
                        frame["__usage"] = self.hoist_item(
                            child["__usage"],
                            frame["__usage"] if "__usage" in frame else {},
                        )

            # if stack is empty, dump the frame
            if len(self.stack) == 0:
                self.write_trace(frame)
            # otherwise, append the frame to the parent
            else:
                if "__frames" not in self.stack[-1]:
                    self.stack[-1]["__frames"] = []
                self.stack[-1]["__frames"].append(frame)

    def hoist_item(self, src: dict[str, Any], cur: dict[str, Any]) -> dict[str, Any]:
        for key, value in src.items():
            if value is None or isinstance(value, list) or isinstance(value, dict):
                continue
            try:
                if key not in cur:
                    cur[key] = value
                else:
                    cur[key] += value
            except Exception:
                # TODO: Be more specific about exceptions here
                continue

        return cur

    def write_trace(self, frame: dict[str, Any]) -> None:
        trace_file = self.output / f"{frame['name']}.{datetime.now().strftime('%Y%m%d.%H%M%S')}.tracy"

        enriched_frame = {
            "runtime": "python",
            "version": VERSION,
            "trace": frame,
        }

        with open(trace_file, "w", encoding="utf-8") as f:
            json.dump(enriched_frame, f, indent=4)


@contextlib.contextmanager
def console_tracer(name: str) -> Iterator[Callable[[str, Any], None]]:
    try:
        print(f"Starting {name}")
        yield lambda key, value: print(f"{key}:\n{json.dumps(to_dict(value), indent=4)}")
    finally:
        print(f"Ending {name}")



# __init__.py
from prompty.invoker import InvokerException

try:
    from .executor import AzureOpenAIExecutor  # noqa
    from .processor import AzureOpenAIProcessor  # noqa
except ImportError:
    raise InvokerException("Error registering AzureOpenAIExecutor and AzureOpenAIProcessor", "azure")



import base64
import re
import typing
from collections.abc import Iterator
from pathlib import Path

import yaml

from .core import Prompty
from .invoker import Parser


class PromptyChatParser(Parser):
    """Prompty Chat Parser"""

    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)
        self.roles = ["assistant", "function", "system", "user", "tools"]
        if isinstance(self.prompty.file, str):
            self.prompty.file = Path(self.prompty.file).resolve().absolute()

        self.path = self.prompty.file.parent

    def inline_image(self, image_item: str) -> str:
        """Inline Image

        Parameters
        ----------
        image_item : str
            The image item to inline

        Returns
        -------
        str
            The inlined image
        """
        # pass through if it's a url or base64 encoded
        if image_item.startswith("http") or image_item.startswith("data"):
            return image_item
        # otherwise, it's a local file - need to base64 encode it
        else:
            image_path = Path(image_item)
            if not image_path.is_absolute():
                image_path = self.path / image_item

            with open(image_path, "rb") as f:
                base64_image = base64.b64encode(f.read()).decode("utf-8")

            if image_path.suffix == ".png":
                return f"data:image/png;base64,{base64_image}"
            elif image_path.suffix == ".jpg":
                return f"data:image/jpeg;base64,{base64_image}"
            elif image_path.suffix == ".jpeg":
                return f"data:image/jpeg;base64,{base64_image}"
            else:
                raise ValueError(
                    f"Invalid image format {image_path.suffix} - currently only .png and .jpg / .jpeg are supported."
                )

    def parse_args(self, args: str) -> dict[str, str]:
        """Parse args

        Parameters
        ----------
        args : str
            The args to parse

        Returns
        -------
        dict[str, str]
            The parsed args
        """

        # regular expression to parse key-value pairs
        string_match = r"\"([^\"]*)\""
        bool_match = r"([Tt]rue|[Ff]alse)"
        float_match = r"([0-9]+(\.[0-9]+))"
        int_match = r"([0-9]+)"

        s = f"({string_match}|{bool_match}|{float_match}|{int_match})"

        patterns = r"(\w+)\s*=\s*(" + s + r")\s*(,?)\s*"

        matches = re.findall(patterns, args)
        full_args = {}
        for m in matches:
            if m[3] != "":
                full_args[m[0]] = m[3]
            elif m[4] != "":
                full_args[m[0]] = m[4].lower() == "true"
            elif m[5] != "":
                full_args[m[0]] = float(m[5])
            elif m[7] != "":
                full_args[m[0]] = int(m[7])

        return full_args

    def parse_content(self, content: str):
        """for parsing inline images

        Parameters
        ----------
        content : str
            The content to parse

        Returns
        -------
        any
            The parsed content
        """
        # regular expression to parse markdown images
        image = r"(?P<alt>!\[[^\]]*\])\((?P<filename>.*?)(?=\"|\))\)"
        matches = re.findall(image, content, flags=re.MULTILINE)
        if len(matches) > 0:
            content_items = []
            content_chunks = re.split(image, content, flags=re.MULTILINE)
            current_chunk = 0
            for i in range(len(content_chunks)):
                # image entry
                if current_chunk < len(matches) and content_chunks[i] == matches[current_chunk][0]:
                    content_items.append(
                        {
                            "type": "image_url",
                            "image_url": {"url": self.inline_image(matches[current_chunk][1].split(" ")[0].strip())},
                        }
                    )
                # second part of image entry
                elif current_chunk < len(matches) and content_chunks[i] == matches[current_chunk][1]:
                    current_chunk += 1
                # text entry
                else:
                    if len(content_chunks[i].strip()) > 0:
                        content_items.append({"type": "text", "text": content_chunks[i].strip()})
            return content_items
        else:
            return content

    def parse(self, data: str) -> Iterator[dict[str, typing.Any]]:
        """Stream the data

        Parameters
        ----------
        data : str
            The data to stream

        Returns
        -------
        Iterator[str]
            The streamed data
        """
        # regular expression to capture boundary roles with optional key-value pairs
        boundary = (
            r"(?i)^\s*#?\s*(" + "|".join(self.roles) + r")(\[((\w+)*\s*=\s*\"?([^\"]*)\"?\s*(,?)\s*)+\])?\s*:\s*$"
        )
        content_buffer: list[str] = []
        # first role is system (if not specified)
        arg_buffer = {"role": "system"}

        for line in data.splitlines():
            # check of ![thread]
            if line.strip().startswith("![thread]"):
                # if content buffer is not empty, then add to messages
                if len(content_buffer) > 0:
                    yield arg_buffer | {"content": "\n".join(content_buffer)}
                    content_buffer = []

                arg_buffer = {
                    "role": "thread",
                }

            # check if line is a boundary
            elif re.match(boundary, line):
                # if content buffer is not empty, then add to messages
                if len(content_buffer) > 0:
                    yield arg_buffer | {"content": "\n".join(content_buffer)}
                    content_buffer = []

                # boundary check for args
                if "[" in line and "]" in line:
                    role, args = line[:-2].split("[", 2)
                    arg_buffer = self.parse_args(args) | {
                        "role": role.strip().lower(),
                    }
                # standard boundary
                else:
                    arg_buffer = {
                        "role": line.replace(":", "").strip().lower(),
                    }
            else:
                content_buffer.append(line)

        # add last message
        if len(content_buffer) > 0:
            yield arg_buffer | {"content": "\n".join(content_buffer)}

    def invoke(self, data: str) -> list[dict[str, str]]:
        """Invoke the Prompty Chat Parser

        Parameters
        ----------
        data : str
            The data to parse

        Returns
        -------
        str
            The parsed data
        """

        messages = []
        for item in self.parse(data):
            if "content" in item:
                item["content"] = self.parse_content(item["content"])

            messages.append(item)

        return messages

    async def invoke_async(self, data: str) -> list[dict[str, str]]:
        """Invoke the Prompty Chat Parser (Async)

        Parameters
        ----------
        data : str
            The data to parse

        Returns
        -------
        str
            The parsed data
        """
        return self.invoke(data)

    def sanitize(self, data):
        # gets template before rendering
        # to clean up any sensitive data
        sanitized_prompt = []
        for item in self.parse(data):
            # add nonce to pre-rendered roles
            item["nonce"] = self.prompty.template.nonce
            role = item.pop("role")
            if "content" not in item:
                item["content"] = ""

            content = item.pop("content")

            # no need to sanitize threads
            if role == "thread":
                sanitized_prompt.append("![thread]")
                sanitized_prompt.append(content)
            else:

                def stringify(x):
                    return f'"{str(x)}"' if isinstance(x, str) else str(x)

                attr = [f"{k}={stringify(v)}" for k, v in item.items()]
                boundary = ",".join(attr)
                sanitized_prompt.append(f"{role}[{boundary}]:")
                sanitized_prompt.append(content)

        return "\n".join(sanitized_prompt)

    def process(self, data):
        # gets template after parse
        # to manage any parsed prompty
        # settings (in  this case, tools)
        if len(data) > 0 and data[0]["role"] == "tools":
            if self.prompty.template.strict and data[0]["nonce"] != self.prompty.template.nonce:
                raise ValueError("Nonce mismatch. Dynamic tools section not allowed in strict mode.")

            content = "tools:\n" + data[0]["content"]
            tools_dict = yaml.load(content, Loader=yaml.FullLoader)
            tools = Prompty.load_tools(tools_dict["tools"])
            self.prompty.merge_tools(tools)

            # remove first item from data
            data = data[1:]

        return data



import typing
from collections.abc import Iterator

import azure.identity
from azure.ai.inference import (
    ChatCompletionsClient,
    EmbeddingsClient,
)
from azure.ai.inference.aio import (
    ChatCompletionsClient as AsyncChatCompletionsClient,
)
from azure.ai.inference.aio import (
    EmbeddingsClient as AsyncEmbeddingsClient,
)
from azure.ai.inference.models import (
    AsyncStreamingChatCompletions,
    StreamingChatCompletions,
)
from azure.core.credentials import AzureKeyCredential

from .._version import VERSION
from ..core import AsyncPromptyStream, Prompty, PromptyStream
from ..invoker import Invoker, InvokerFactory
from ..tracer import Tracer


@InvokerFactory.register_executor("serverless")
class ServerlessExecutor(Invoker):
    """Azure OpenAI Executor"""

    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)
        self.kwargs = {key: value for key, value in self.prompty.model.connection.items() if key != "type"}

        self.endpoint = self.prompty.model.connection["endpoint"]
        self.model = self.prompty.model.connection["model"]

        # no key, use default credentials
        if "key" not in self.kwargs:
            self.credential: typing.Union[azure.identity.DefaultAzureCredential, AzureKeyCredential] = (
                azure.identity.DefaultAzureCredential(exclude_shared_token_cache_credential=True)
            )
        else:
            self.credential = AzureKeyCredential(self.prompty.model.connection["key"])

        # api type
        self.api = self.prompty.model.api

    def _response(self, response: typing.Any) -> typing.Any:
        # stream response
        if isinstance(response, Iterator):
            if isinstance(response, StreamingChatCompletions):
                stream: typing.Union[PromptyStream, AsyncPromptyStream] = PromptyStream("ServerlessExecutor", response)
                return stream
            elif isinstance(response, AsyncStreamingChatCompletions):
                stream = AsyncPromptyStream("ServerlessExecutor", response)
                return stream
            else:
                stream = PromptyStream("ServerlessExecutor", response)

            return stream
        else:
            return response

    def _sanitize_messages(self, data: typing.Any) -> list[dict[str, str]]:
        messages = data if isinstance(data, list) else [data]

        if self.prompty.template.strict:
            if not all([msg["nonce"] == self.prompty.template.nonce for msg in messages]):
                raise ValueError("Nonce mismatch in messages array (strict mode)")

        messages = [
            {
                **{"role": msg["role"], "content": msg["content"]},
                **({"name": msg["name"]} if "name" in msg else {}),
            }
            for msg in messages
        ]

        return messages

    def invoke(self, data: typing.Any) -> typing.Any:
        """Invoke the Serverless SDK

        Parameters
        ----------
        data : any
            The data to send to the Serverless SDK

        Returns
        -------
        any
            The response from the Serverless SDK
        """

        cargs = {
            "endpoint": self.endpoint,
            "credential": self.credential,
        }

        if self.api == "chat":
            with Tracer.start("ChatCompletionsClient") as trace:
                trace("type", "LLM")
                trace("signature", "azure.ai.inference.ChatCompletionsClient.ctor")
                trace("description", "Azure Unified Inference SDK Chat Completions Client")
                trace("inputs", cargs)
                client: typing.Any = ChatCompletionsClient(
                    user_agent=f"prompty/{VERSION}",
                    **cargs,
                )
                trace("result", client)

            with Tracer.start("complete") as trace:
                trace("type", "LLM")
                trace("signature", "azure.ai.inference.ChatCompletionsClient.complete")
                trace("description", "Azure Unified Inference SDK Chat Completions Client")
                eargs = {
                    "model": self.model,
                    "messages": self._sanitize_messages(data),
                    **self.prompty.model.options,
                }
                trace("inputs", eargs)
                r = client.complete(**eargs)
                trace("result", r)

            response = self._response(r)

        elif self.api == "completion":
            raise NotImplementedError("Serverless Completions API is not implemented yet")

        elif self.api == "embedding":
            with Tracer.start("EmbeddingsClient") as trace:
                trace("type", "LLM")
                trace("signature", "azure.ai.inference.EmbeddingsClient.ctor")
                trace("description", "Azure Unified Inference SDK Embeddings Client")
                trace("inputs", cargs)
                client = EmbeddingsClient(
                    user_agent=f"prompty/{VERSION}",
                    **cargs,
                )
                trace("result", client)

            with Tracer.start("complete") as trace:
                trace("type", "LLM")
                trace("signature", "azure.ai.inference.ChatCompletionsClient.complete")
                trace("description", "Azure Unified Inference SDK Chat Completions Client")
                eargs = {
                    "model": self.model,
                    "input": data if isinstance(data, list) else [data],
                    **self.prompty.model.options,
                }
                trace("inputs", eargs)
                r = client.complete(**eargs)
                trace("result", r)

            response = self._response(r)

        elif self.api == "image":
            raise NotImplementedError("Azure OpenAI Image API is not implemented yet")

        return response

    async def invoke_async(self, data: str) -> str:
        """Invoke the Prompty Chat Parser (Async)

        Parameters
        ----------
        data : str
            The data to parse

        Returns
        -------
        str
            The parsed data
        """
        cargs = {
            "endpoint": self.endpoint,
            "credential": self.credential,
        }

        if self.api == "chat":
            with Tracer.start("ChatCompletionsClient") as trace:
                trace("type", "LLM")
                trace("signature", "azure.ai.inference.aio.ChatCompletionsClient.ctor")
                trace(
                    "description",
                    "Azure Unified Inference SDK Async Chat Completions Client",
                )
                trace("inputs", cargs)
                client: typing.Any = AsyncChatCompletionsClient(
                    user_agent=f"prompty/{VERSION}",
                    **cargs,
                )
                trace("result", client)

            with Tracer.start("complete") as trace:
                trace("type", "LLM")
                trace("signature", "azure.ai.inference.ChatCompletionsClient.complete")
                trace(
                    "description",
                    "Azure Unified Inference SDK Async Chat Completions Client",
                )
                eargs = {
                    "model": self.model,
                    "messages": self._sanitize_messages(data),
                    **self.prompty.model.options,
                }
                trace("inputs", eargs)
                r = await client.complete(**eargs)
                trace("result", r)

            response = self._response(r)

        elif self.api == "completion":
            raise NotImplementedError("Serverless Completions API is not implemented yet")

        elif self.api == "embedding":
            with Tracer.start("EmbeddingsClient") as trace:
                trace("type", "LLM")
                trace("signature", "azure.ai.inference.aio.EmbeddingsClient.ctor")
                trace("description", "Azure Unified Inference SDK Async Embeddings Client")
                trace("inputs", cargs)
                client = AsyncEmbeddingsClient(
                    user_agent=f"prompty/{VERSION}",
                    **cargs,
                )
                trace("result", client)

            with Tracer.start("complete") as trace:
                trace("type", "LLM")
                trace("signature", "azure.ai.inference.ChatCompletionsClient.complete")
                trace("description", "Azure Unified Inference SDK Chat Completions Client")
                eargs = {
                    "model": self.model,
                    "input": data if isinstance(data, list) else [data],
                    **self.prompty.model.options,
                }
                trace("inputs", eargs)
                r = await client.complete(**eargs)
                trace("result", r)

            response = self._response(r)

        elif self.api == "image":
            raise NotImplementedError("Azure OpenAI Image API is not implemented yet")

        return response



import json
import re
import typing
from pathlib import Path

import aiofiles
import yaml

_yaml_regex = re.compile(
    r"^\s*" + r"(?:---|\+\+\+)" + r"(.*?)" + r"(?:---|\+\+\+)" + r"\s*(.+)$",
    re.S | re.M,
)


def load_text(file_path, encoding="utf-8"):
    with open(file_path, encoding=encoding) as file:
        return file.read()


async def load_text_async(file_path, encoding="utf-8"):
    async with aiofiles.open(file_path, encoding=encoding) as f:
        content = await f.read()
        return content


def load_json(file_path, encoding="utf-8"):
    return json.loads(load_text(file_path, encoding=encoding))


async def load_json_async(file_path, encoding="utf-8"):
    # async file open
    content = await load_text_async(file_path, encoding=encoding)
    return json.loads(content)


def _walk_up_path(path: Path) -> typing.Union[Path, None]:
    """Walk up the path to find a prompty.json file.

    Args:
        path (Path): The path to start searching from.
    """
    while path != path.parent:
        # Check if the prompty.json file exists in the current directory
        prompty_config = path / "prompty.json"
        if prompty_config.exists():
            return prompty_config
        # Move up to the parent directory
        path = path.parent
    return None


def _find_global_config(prompty_path: Path = Path.cwd()) -> typing.Union[Path, None]:
    """Find the prompty.json file in the current directory or any parent directory.
    Args:
        prompty_path (Path): The path to start searching from.
    """
    if Path(prompty_path / "prompty.json").exists():
        return Path(prompty_path / "prompty.json")
    else:
        return _walk_up_path(prompty_path)


def load_global_config(prompty_path: Path = Path.cwd(), configuration: str = "default") -> dict[str, typing.Any]:
    # prompty.config laying around?
    config = _find_global_config(prompty_path)

    # if there is one load it
    if config is not None:
        c = load_json(config)
        if configuration in c:
            return c[configuration]
        else:
            raise ValueError(f'Item "{configuration}" not found in "{config}"')

    return {}


async def load_global_config_async(
    prompty_path: Path = Path.cwd(), configuration: str = "default"
) -> dict[str, typing.Any]:
    # prompty.config laying around?
    config = _find_global_config(prompty_path)

    # if there is one load it
    if config is not None:
        c = await load_json_async(config)
        if configuration in c:
            return c[configuration]
        else:
            raise ValueError(f'Item "{configuration}" not found in "{config}"')

    return {}


def get_json_type(t: type) -> typing.Literal["string", "number", "array", "object", "boolean"]:
    if t is str:
        return "string"
    elif t is int:
        return "number"
    elif t is float:
        return "number"
    elif t is list:
        return "array"
    elif t is dict:
        return "object"
    elif t is bool:
        return "boolean"
    else:
        raise ValueError(f"Unsupported type: {t}")


def load_prompty(file_path, encoding="utf-8"):
    contents = load_text(file_path, encoding=encoding)
    return parse(contents)


async def load_prompty_async(file_path, encoding="utf-8"):
    contents = await load_text_async(file_path, encoding=encoding)
    return parse(contents)


def parse(contents):
    global _yaml_regex

    fmatter = ""
    body = ""
    result = _yaml_regex.search(contents)

    if result:
        fmatter = result.group(1)
        body = result.group(2)
    return {
        "attributes": yaml.load(fmatter, Loader=yaml.FullLoader),
        "body": body,
        "frontmatter": fmatter,
    }



import json
import typing
from collections.abc import AsyncIterator, Iterator

import azure.identity
from openai import APIResponse, AsyncAzureOpenAI, AzureOpenAI
from openai.types.chat.chat_completion import ChatCompletion

from prompty.tracer import Tracer

from .._version import VERSION
from ..common import convert_function_tools, convert_output_props
from ..core import AsyncPromptyStream, Prompty, PromptyStream
from ..invoker import Invoker, InvokerFactory


@InvokerFactory.register_executor("azure")
@InvokerFactory.register_executor("azure_openai")
class AzureOpenAIExecutor(Invoker):
    """Azure OpenAI Executor"""

    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)

        # resolve model connection and options
        self.resolve_model()

        self.kwargs = {key: value for key, value in self.prompty.model.connection.items() if key != "type"}

        # no key, use default credentials
        if "api_key" not in self.kwargs:
            # managed identity if client id
            if "client_id" in self.kwargs:
                default_credential: typing.Union[
                    azure.identity.ManagedIdentityCredential,
                    azure.identity.DefaultAzureCredential,
                ] = azure.identity.ManagedIdentityCredential(
                    client_id=self.kwargs.pop("client_id"),
                )
            # default credential
            else:
                default_credential = azure.identity.DefaultAzureCredential(exclude_shared_token_cache_credential=True)

            self.kwargs["azure_ad_token_provider"] = azure.identity.get_bearer_token_provider(
                default_credential, "https://cognitiveservices.azure.com/.default"
            )

        self.api = self.prompty.model.api
        self.deployment = self.prompty.model.connection["azure_deployment"]
        self.options = self.prompty.model.options

    def _sanitize_messages(self, data: typing.Any, ignore_thread_content=False) -> list[dict[str, str]]:
        messages = data if isinstance(data, list) else [data]

        if self.prompty.template.strict:
            if not all([msg["nonce"] == self.prompty.template.nonce for msg in messages]):
                raise ValueError("Nonce mismatch in messages array (strict mode)")

        messages = []
        for msg in data:
            if msg["role"] == "thread":
                thread = self.prompty.get_input("thread")
                if thread is None:
                    raise ValueError("thread requires thread input")

                if thread.value is None:
                    thread.value = []

                if not ignore_thread_content:
                    thread.value.append({"role": "user", "content": msg["content"]})

                if isinstance(thread.value, list):
                    messages = [*messages, *thread.value]
                elif isinstance(thread.value, dict):
                    messages.append(
                        {
                            **{"role": thread.value["role"], "content": thread.value["content"]},
                            **({"name": thread.value["name"]} if "name" in thread.value else {}),
                        }
                    )
                else:
                    messages.append(thread.value)
            else:
                messages.append(
                    {
                        **{"role": msg["role"], "content": msg["content"]},
                        **({"name": msg["name"]} if "name" in msg else {}),
                    }
                )

        return messages

    def _get_ctor(self) -> AzureOpenAI:
        with Tracer.start("AzureOpenAI") as trace:
            trace("type", "LLM")
            trace("signature", "AzureOpenAI.ctor")
            trace("description", "Azure OpenAI Constructor")
            trace("inputs", self.kwargs)
            client = AzureOpenAI(
                default_headers={
                    "User-Agent": f"prompty/{VERSION}",
                    "x-ms-useragent": f"prompty/{VERSION}",
                },
                **self.kwargs,
            )
            trace("result", client)
            return client

    def _get_async_ctor(self) -> AsyncAzureOpenAI:
        with Tracer.start("AzureOpenAIAsync") as trace:
            trace("type", "LLM")
            trace("signature", "AzureOpenAIAsync.ctor")
            trace("description", "Async Azure OpenAI Constructor")
            trace("inputs", self.kwargs)
            client = AsyncAzureOpenAI(
                default_headers={
                    "User-Agent": f"prompty/{VERSION}",
                    "x-ms-useragent": f"prompty/{VERSION}",
                },
                **self.kwargs,
            )
            trace("result", client)
            return client

    def _resolve_chat_args(self, data: typing.Any, ignore_thread_content=False) -> dict:
        messages = self._sanitize_messages(data, ignore_thread_content)

        args = {
            "model": self.deployment,
            "messages": messages,
            **self.options,
        }

        if "tools" not in self.options and len(self.prompty.tools) > 0:
            # add tools to options:
            args = {**args, "tools": convert_function_tools(self.prompty.tools)}

        if len(self.prompty.outputs) > 0:
            # add outputs to options:
            args = {
                **args,
                "response_format": convert_output_props(
                    self.prompty.name.lower().replace(" ", "_"), self.prompty.outputs
                ),
            }

        return args

    def _create_chat(self, client: AzureOpenAI, data: typing.Any, ignore_thread_content=False) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")
            trace("signature", "AzureOpenAI.chat.completions.create")
            args = self._resolve_chat_args(data, ignore_thread_content)
            trace("inputs", args)
            if "stream" in args and args["stream"]:
                response = client.chat.completions.create(**args)
            else:
                raw = client.chat.completions.with_raw_response.create(**args)

                response = ChatCompletion.model_validate_json(raw.text)

                for k, v in raw.headers.raw:
                    trace(k.decode("utf-8"), v.decode("utf-8"))

                trace("request_id", raw.request_id)
                trace("retries_taken", raw.retries_taken)
            trace("result", response)
            return response

    async def _create_chat_async(
        self, client: AsyncAzureOpenAI, data: typing.Any, ignore_thread_content=False
    ) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            trace("signature", "AzureOpenAIAsync.chat.completions.create")
            args = self._resolve_chat_args(data, ignore_thread_content)
            trace("inputs", args)
            if "stream" in args and args["stream"]:
                response = await client.chat.completions.create(**args)
            else:
                raw: APIResponse = await client.chat.completions.with_raw_response.create(**args)
                if raw is not None and raw.text is not None and isinstance(raw.text, str):
                    response = ChatCompletion.model_validate_json(raw.text)

                for k, v in raw.headers.raw:
                    trace(k.decode("utf-8"), v.decode("utf-8"))

                trace("request_id", raw.request_id)
                trace("retries_taken", raw.retries_taken)
            trace("result", response)

            return response

    def _execute_agent(self, client: AzureOpenAI, data: typing.Any) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            trace("signature", "AzureOpenAI.chat.agent.create")
            trace("inputs", data)

            response = self._create_chat(client, data)
            if isinstance(response, ChatCompletion):
                message = response.choices[0].message
                if message.tool_calls:
                    thread = self.prompty.get_input("thread")
                    if thread is None:
                        raise ValueError("thread requires thread input")

                    thread.value.append(
                        {
                            "role": "assistant",
                            "tool_calls": [t.model_dump() for t in message.tool_calls],
                        }
                    )

                    for tool_call in message.tool_calls:
                        tool = self.prompty.get_tool(tool_call.function.name)
                        if tool is None:
                            raise ValueError(f"Tool {tool_call.function.name} does not exist")

                        function_args = json.loads(tool_call.function.arguments)

                        if tool.value is None:
                            raise ValueError(f"Tool {tool_call.function.name} does not have a value")

                        r = tool.value(**function_args)

                        thread.value.append(
                            {
                                "role": "tool",
                                "tool_call_id": tool_call.id,
                                "name": tool_call.function.name,
                                "content": r,
                            }
                        )
                else:
                    trace("result", response)
                    return response

            response = self._create_chat(client, data, True)
            trace("result", response)

            return response

    async def _execute_agent_async(self, client: AsyncAzureOpenAI, data: typing.Any) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")
            trace("signature", "AzureOpenAI.chat.agent.create")
            args = self._resolve_chat_args(data)
            trace("inputs", args)
            response = 5
            trace("result", response)
            return response

    def _create_completion(self, client: AzureOpenAI, data: typing.Any) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            trace("signature", "AzureOpenAI.completions.create")
            args = {
                "prompt": data,
                "model": self.deployment,
                **self.options,
            }
            trace("inputs", args)
            response = client.completions.create(**args)
            trace("result", response)
            return response

    async def _create_completion_async(self, client: AsyncAzureOpenAI, data: typing.Any) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            trace("signature", "AzureOpenAIAsync.completions.create")
            args = {
                "prompt": data,
                "model": self.deployment,
                **self.options,
            }
            trace("inputs", args)

            response = await client.completions.create(**args)
            trace("result", response)

            return response

    def _create_embedding(self, client: AzureOpenAI, data: typing.Any) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            trace("signature", "AzureOpenAI.embeddings.create")
            args = {
                "input": data if isinstance(data, list) else [data],
                "model": self.deployment,
                **self.options,
            }
            trace("inputs", args)
            response = client.embeddings.create(**args)
            trace("result", response)

            return response

    async def _create_embedding_async(self, client: AsyncAzureOpenAI, data: typing.Any) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            trace("signature", "AzureOpenAIAsync.embeddings.create")
            args = {
                "input": data if isinstance(data, list) else [data],
                "model": self.deployment,
                **self.options,
            }
            trace("inputs", args)
            response = await client.embeddings.create(**args)
            trace("result", response)

            return response

    def _create_image(self, client: AzureOpenAI, data: typing.Any) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            trace("signature", "AzureOpenAI.images.generate")
            args = {
                "prompt": data,
                "model": self.deployment,
                **self.options,
            }
            trace("inputs", args)
            response = client.images.generate(**args)
            trace("result", response)
            return response

    async def _create_image_async(self, client: AsyncAzureOpenAI, data: typing.Any) -> typing.Any:
        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            trace("signature", "AzureOpenAIAsync.images.generate")
            args = {
                "prompt": data,
                "model": self.deployment,
                **self.options,
            }
            trace("inputs", args)
            response = await client.images.generate(**args)
            trace("result", response)

            return response

    def invoke(self, data: typing.Any) -> typing.Union[str, PromptyStream]:
        """Invoke the Azure OpenAI API

        Parameters
        ----------
        data : any
            The data to send to the Azure OpenAI API

        Returns
        -------
        any
            The response from the Azure OpenAI API
        """

        client = self._get_ctor()

        if self.api == "chat":
            response = self._create_chat(client, data)
        elif self.api == "agent":
            response = self._execute_agent(client, data)
        elif self.api == "completion":
            response = self._create_completion(client, data)
        elif self.api == "embedding":
            response = self._create_embedding(client, data)
        elif self.api == "image":
            response = self._create_image(client, data)

        # stream response
        if isinstance(response, Iterator):
            if self.api == "chat":
                # TODO: handle the case where there might be no usage in the stream
                return PromptyStream("AzureOpenAIExecutor", response)
            else:
                return PromptyStream("AzureOpenAIExecutor", response)
        else:
            return response

    async def invoke_async(self, data: str) -> typing.Union[str, AsyncPromptyStream]:
        """Invoke the Prompty Chat Parser (Async)

        Parameters
        ----------
        data : str
            The data to parse

        Returns
        -------
        str
            The parsed data
        """
        client = self._get_async_ctor()

        if self.api == "chat":
            response = await self._create_chat_async(client, data)
        elif self.api == "agent":
            response = await self._execute_agent_async(client, data)
        elif self.api == "completion":
            response = await self._create_completion_async(client, data)
        elif self.api == "embedding":
            response = await self._create_embedding_async(client, data)
        elif self.api == "image":
            response = await self._create_image_async(client, data)

        # stream response
        if isinstance(response, AsyncIterator):
            if self.api == "chat":
                # TODO: handle the case where there might be no usage in the stream
                return AsyncPromptyStream("AzureOpenAIExecutorAsync", response)
            else:
                return AsyncPromptyStream("AzureOpenAIExecutorAsync", response)
        else:
            return response



import pytest

import prompty


@pytest.mark.parametrize(
    "prompt",
    [
        "mustache/basic.prompty",
    ],
)
def test_load(prompt: str):
    p = prompty.load(prompt)
    o = prompty.prepare(
        p,
        inputs={
            "firstName": "John",
            "lastName": "Doe",
            "question": "What is your name?",
        },
    )
    print(o)



import re
import typing
from collections.abc import AsyncIterator, Iterator
from datetime import datetime

import azure.identity
from openai import AsyncAzureOpenAI, AzureOpenAI

from prompty.tracer import Tracer

from .._version import VERSION
from ..core import AsyncPromptyStream, Prompty, PromptyStream
from ..invoker import Invoker, InvokerFactory


def extract_date(data: str) -> typing.Union[datetime, None]:
    """Extract date from a string

    Parameters
    ----------
    data : str
        The string containing the date

    Returns
    -------
    datetime
        The extracted date as a datetime object
    """

    # Regular expression to find dates in the format YYYY-MM-DD
    date_pattern = re.compile(r"\b\d{4}-\d{2}-\d{2}\b")
    match = date_pattern.search(data)
    if match:
        date_str = match.group(0)
        # Validate the date format
        try:
            return datetime.strptime(date_str, "%Y-%m-%d")
        except ValueError:
            pass
    return None


def is_structured_output_available(api_version: str) -> bool:
    """Check if the structured output API is available for the given API version

    Parameters
    ----------
    api_version : datetime
        The API version

    Returns
    -------
    bool
        True if the structured output API is available, False otherwise
    """

    # Define the threshold date
    threshold_api_version_date = datetime(2024, 8, 1)

    api_version_date = extract_date(api_version)

    # Check if the API version are on or after the threshold date
    if api_version_date is not None and api_version_date >= threshold_api_version_date:
        return True
    return False


@InvokerFactory.register_executor("azure_beta")
@InvokerFactory.register_executor("azure_openai_beta")
class AzureOpenAIBetaExecutor(Invoker):
    """Azure OpenAI Beta Executor"""

    def __init__(self, prompty: Prompty) -> None:
        super().__init__(prompty)
        self.kwargs = {key: value for key, value in self.prompty.model.connection.items() if key != "type"}

        # no key, use default credentials
        if "api_key" not in self.kwargs:
            # managed identity if client id
            if "client_id" in self.kwargs:
                default_credential: typing.Union[
                    azure.identity.ManagedIdentityCredential,
                    azure.identity.DefaultAzureCredential,
                ] = azure.identity.ManagedIdentityCredential(
                    client_id=self.kwargs.pop("client_id"),
                )
            # default credential
            else:
                default_credential = azure.identity.DefaultAzureCredential(exclude_shared_token_cache_credential=True)

            self.kwargs["azure_ad_token_provider"] = azure.identity.get_bearer_token_provider(
                default_credential, "https://cognitiveservices.azure.com/.default"
            )

        self.api = self.prompty.model.api
        self.api_version = self.prompty.model.connection["api_version"]
        self.deployment = self.prompty.model.connection["azure_deployment"]
        self.options = self.prompty.model.options

    def _sanitize_messages(self, data: typing.Any) -> list[dict[str, str]]:
        messages = data if isinstance(data, list) else [data]

        if self.prompty.template.strict:
            if not all([msg["nonce"] == self.prompty.template.nonce for msg in messages]):
                raise ValueError("Nonce mismatch in messages array (strict mode)")

        messages = [
            {
                **{"role": msg["role"], "content": msg["content"]},
                **({"name": msg["name"]} if "name" in msg else {}),
            }
            for msg in messages
        ]

        return messages

    def invoke(self, data: typing.Any) -> typing.Any:
        """Invoke the Azure OpenAI API

        Parameters
        ----------
        data : any
            The data to send to the Azure OpenAI API

        Returns
        -------
        any
            The response from the Azure OpenAI API
        """

        with Tracer.start("AzureOpenAI") as trace:
            trace("type", "LLM")
            trace("signature", "AzureOpenAI.ctor")
            trace("description", "Azure OpenAI Constructor")
            trace("inputs", self.kwargs)
            client = AzureOpenAI(
                default_headers={
                    "User-Agent": f"prompty/{VERSION}",
                    "x-ms-useragent": f"prompty/{VERSION}",
                },
                **self.kwargs,
            )
            trace("result", client)

        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            if self.api == "chat":
                # We can only verify the API version as the model and its version are not part of prompty configuration
                # Should be gpt-4o and 2024-08-06 or later
                choose_beta = is_structured_output_available(self.api_version)
                if choose_beta:
                    trace("signature", "AzureOpenAI.beta.chat.completions.parse")
                else:
                    trace("signature", "AzureOpenAI.chat.completions.create")

                args = {
                    "model": self.deployment,
                    "messages": data if isinstance(data, list) else [data],
                    **self.options,
                }
                trace("inputs", args)
                if choose_beta:
                    response: typing.Any = client.beta.chat.completions.parse(**args)
                else:
                    response = client.chat.completions.create(**args)
                trace("result", response)

            elif self.api == "completion":
                trace("signature", "AzureOpenAI.completions.create")
                args = {
                    "prompt": data,
                    "model": self.deployment,
                    **self.options,
                }
                trace("inputs", args)
                response = client.completions.create(**args)
                trace("result", response)

            elif self.api == "embedding":
                trace("signature", "AzureOpenAI.embeddings.create")
                args = {
                    "input": data if isinstance(data, list) else [data],
                    "model": self.deployment,
                    **self.options,
                }
                trace("inputs", args)
                response = client.embeddings.create(**args)
                trace("result", response)

            elif self.api == "image":
                trace("signature", "AzureOpenAI.images.generate")
                args = {
                    "prompt": data,
                    "model": self.deployment,
                    **self.options,
                }
                trace("inputs", args)
                response = client.images.generate(**args)
                trace("result", response)

        # stream response
        if isinstance(response, Iterator):
            if self.api == "chat":
                # TODO: handle the case where there might be no usage in the stream
                return PromptyStream("AzureOpenAIBetaExecutor", response)
            else:
                return PromptyStream("AzureOpenAIBetaExecutor", response)
        else:
            return response

    async def invoke_async(self, data: str) -> typing.Union[str, AsyncPromptyStream]:
        """Invoke the Prompty Chat Parser (Async)

        Parameters
        ----------
        data : str
            The data to parse

        Returns
        -------
        str
            The parsed data
        """
        with Tracer.start("AzureOpenAIAsync") as trace:
            trace("type", "LLM")
            trace("signature", "AzureOpenAIAsync.ctor")
            trace("description", "Async Azure OpenAI Constructor")
            trace("inputs", self.kwargs)
            client = AsyncAzureOpenAI(
                default_headers={
                    "User-Agent": f"prompty/{VERSION}",
                    "x-ms-useragent": f"prompty/{VERSION}",
                },
                **self.kwargs,
            )
            trace("result", client)

        with Tracer.start("create") as trace:
            trace("type", "LLM")
            trace("description", "Azure OpenAI Client")

            if self.api == "chat":
                trace("signature", "AzureOpenAIAsync.chat.completions.create")
                args = {
                    "model": self.deployment,
                    "messages": self._sanitize_messages(data),
                    **self.options,
                }
                trace("inputs", args)
                response = await client.chat.completions.create(**args)
                trace("result", response)

            elif self.api == "completion":
                trace("signature", "AzureOpenAIAsync.completions.create")
                args = {
                    "prompt": data,
                    "model": self.deployment,
                    **self.options,
                }
                trace("inputs", args)
                response = await client.completions.create(**args)
                trace("result", response)

            elif self.api == "embedding":
                trace("signature", "AzureOpenAIAsync.embeddings.create")
                args = {
                    "input": data if isinstance(data, list) else [data],
                    "model": self.deployment,
                    **self.options,
                }
                trace("inputs", args)
                response = await client.embeddings.create(**args)
                trace("result", response)

            elif self.api == "image":
                trace("signature", "AzureOpenAIAsync.images.generate")
                args = {
                    "prompt": data,
                    "model": self.deployment,
                    **self.options,
                }
                trace("inputs", args)
                response = await client.images.generate(**args)
                trace("result", response)

        # stream response
        if isinstance(response, AsyncIterator):
            if self.api == "chat":
                # TODO: handle the case where there might be no usage in the stream
                return AsyncPromptyStream("AzureOpenAIBetaExecutorAsync", response)
            else:
                return AsyncPromptyStream("AzureOpenAIBetaExecutorAsync", response)
        else:
            return response



import typing

from .core import OutputProperty, ToolProperty


def convert_function_tools(tools: list[ToolProperty]) -> list[dict[str, typing.Any]]:
    """Convert the tools to a list of dictionaries
    Parameters
    ----------
    tools : list[ToolProperty]
        The tools to convert
    Returns
    -------
    list[dict[str, typing.Any]]
        The converted tools
    """
    if tools:
        return [
            {
                "type": "function",
                "function": {
                    "name": tool.id,
                    "description": tool.description,
                    "parameters": {
                        "type": "object",
                        "properties": {
                            p.name: {
                                "type": p.type,
                                **({"description": p.description} if p.description else {}),
                                **({"enum": p.enum} if p.enum else {}),
                            }
                            for p in tool.parameters
                        },
                        "required": [p.name for p in tool.parameters if p.required],
                    },
                },
            }
            for tool in tools
            if tool.type == "function"
        ]
    return []


def convert_output_props(name: str, outputs: list[OutputProperty]) -> dict[str, typing.Any]:
    """Convert the tools to a list of dictionaries
    Parameters
    ----------
    tools : list[OutputProperty]
        The tools to convert
    Returns
    -------
    list[dict[str, typing.Any]]
        The converted tools
    """
    if outputs:
        return {
            "type": "json_schema",
            "json_schema": {
                "name": name,
                "strict": True,
                "schema": {
                    "type": "object",
                    "properties": {
                        p.name: {
                            "type": p.type,
                        }
                        for p in outputs
                    },
                    "required": [p.name for p in outputs],
                    "additionalProperties": False,
                },
            },
        }
    return {}
